# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provided interface to read inheritance statements from Markdown."""

import logging
import os
import pathlib
from typing import Dict, List, Tuple

import trestle.core.generators as gens
import trestle.oscal.ssp as ossp
from trestle.common.common_types import TypeWithByComps
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_list, none_if_empty
from trestle.core.crm.bycomp_interface import ByComponentInterface
from trestle.core.crm.leveraged_statements import InheritanceMarkdownReader

logger = logging.getLogger(__name__)

# Provide name for this type
# Containing dictionary that is keyed by by_component uuid with a tuple of inherited and satisfied statements
ByComponentDict = Dict[str, Tuple[List[ossp.Inherited], List[ossp.Satisfied]]]

# Provide name for this type
# Containing dictionary that is keyed by control id with a dictionary of by_component information
InheritanceViewDict = Dict[str, ByComponentDict]


class ExportReader:
    """
    By-Component Assembly Exports Markdown reader.

    Export reader handles all operations related to reading authored inherited and satisfied statements from exports
    in Markdown. The reader will read all the markdown files in the exports directory and update the SSP with the
    inheritance.
    """

    def __init__(self, root_path: pathlib.Path, ssp: ossp.SystemSecurityPlan):
        """
        Initialize export reader.

        Arguments:
            root_path: A root path object where an SSP's inheritance markdown is located.
            ssp: A system security plan object that will be updated with the inheritance information.

        Notes:
            The mapped components list is used to track which components have been mapped to controls in the markdown.
            It can be retrieved with the get_leveraged_components method. This will be empty until the
            read_exports_from_markdown method is called.
        """
        self._ssp: ossp.SystemSecurityPlan = ssp

        # Create a dictionary of implemented requirements keyed by control id for merging operations
        self._implemented_requirements: Dict[str, ossp.ImplementedRequirement] = self._create_impl_req_dict()

        # List of component titles that have been mapped to controls in the Markdown
        self._mapped_components: List[str] = []

        self._root_path: pathlib.Path = root_path

    def _create_impl_req_dict(self) -> Dict[str, ossp.ImplementedRequirement]:
        """Create a dictionary of implemented requirements keyed by control id."""
        impl_req_dict: Dict[str, ossp.ImplementedRequirement] = {}
        for impl_req in as_list(self._ssp.control_implementation.implemented_requirements):
            impl_req_dict[impl_req.control_id] = impl_req
        return impl_req_dict

    def read_exports_from_markdown(self) -> ossp.SystemSecurityPlan:
        """Read inheritance markdown and update the SSP with the inheritance information."""
        # Read the information from the markdown files into a dictionary for quick lookup
        markdown_dict: InheritanceViewDict = self._read_inheritance_markdown_directory()

        # Merge the markdown information into existing the implemented requirements
        self._merge_exports_implemented_requirements(markdown_dict)

        # Process remaining markdown information that was not in the implemented requirements
        for control_id, by_comp_dict in markdown_dict.items():
            if by_comp_dict:
                logging.debug(f'Adding control mapping {control_id} to implemented requirements')
                self._add_control_mappings_to_implemented_requirements(control_id, by_comp_dict)

        self._ssp.control_implementation.implemented_requirements = list(self._implemented_requirements.values())
        return self._ssp

    def get_leveraged_ssp_href(self) -> str:
        """Get the href of the leveraged SSP from a markdown file."""
        comp_dirs = os.listdir(self._root_path)
        if len(comp_dirs) == 0:
            raise TrestleError('No components were found in the markdown.')

        control_dirs = os.listdir(self._root_path.joinpath(comp_dirs[0]))
        if len(control_dirs) == 0:
            raise TrestleError('No controls were found in the markdown for component {comp_dirs[0]}.')

        control_dir = self._root_path.joinpath(comp_dirs[0], control_dirs[0])

        files = [f for f in os.listdir(control_dir) if os.path.isfile(os.path.join(control_dir, f))]
        if len(files) == 0:
            raise TrestleError(f'No files were found in the markdown for control {control_dirs[0]}.')

        markdown_file_path = control_dir.joinpath(files[0])
        reader = InheritanceMarkdownReader(markdown_file_path)
        return reader.get_leveraged_ssp_reference()

    def get_leveraged_components(self) -> List[str]:
        """Get a list of component titles that have been mapped to controls in the Markdown."""
        return self._mapped_components

    def _merge_exports_implemented_requirements(self, markdown_dict: InheritanceViewDict) -> None:
        """Merge all exported inheritance info from the markdown into the implemented requirement dict."""
        for implemented_requirement in self._implemented_requirements.values():

            # If the control id existing in the markdown, then update the by_components
            if implemented_requirement.control_id in markdown_dict:

                # Delete the entry from the markdown_dict once processed to avoid duplicates
                by_comp_dict: ByComponentDict = markdown_dict.pop(implemented_requirement.control_id)

                self._update_type_with_by_comp(implemented_requirement, by_comp_dict)

            # Update any implemented requirements statements assemblies
            new_statements: List[ossp.Statement] = []

            for stm in as_list(implemented_requirement.statements):
                statement_id = getattr(stm, 'statement_id', f'{implemented_requirement.control_id}_smt')

                # If the statement id existing in the markdown, then update the by_components
                if statement_id in markdown_dict:

                    # Delete the entry from the markdown_dict once processed to avoid duplicates
                    by_comp_dict: ByComponentDict = markdown_dict.pop(statement_id)

                    self._update_type_with_by_comp(stm, by_comp_dict)

                new_statements.append(stm)

            implemented_requirement.statements = none_if_empty(new_statements)

    def _update_type_with_by_comp(self, with_bycomp: TypeWithByComps, by_comp_dict: ByComponentDict) -> None:
        """Update the by_components for a type with by_components."""
        new_by_comp: List[ossp.ByComponent] = []

        by_comp: ossp.ByComponent
        comp_inheritance_info: Tuple[List[ossp.Inherited], List[ossp.Satisfied]]
        for by_comp in as_list(with_bycomp.by_components):

            # If the by_component uuid exists in the by_comp_dict, then update it
            # If not, clear the by_component inheritance information
            comp_inheritance_info = by_comp_dict.pop(by_comp.component_uuid, ([], []))

            bycomp_interface = ByComponentInterface(by_comp)
            by_comp = bycomp_interface.reconcile_inheritance_by_component(
                comp_inheritance_info[0], comp_inheritance_info[1]
            )

            new_by_comp.append(by_comp)

        # Add any new by_components that were not in the original statement
        new_by_comp.extend(ExportReader._add_new_by_comps(by_comp_dict))
        with_bycomp.by_components = none_if_empty(new_by_comp)

    def _add_control_mappings_to_implemented_requirements(
        self, control_mapping: str, by_comps: ByComponentDict
    ) -> None:
        """Add control mappings to implemented requirements."""
        # Determine if the control id is actually a statement id
        if '_smt.' in control_mapping:
            control_id = control_mapping.split('_smt')[0]
            implemented_req = self._add_or_get_implemented_requirement(control_id)
            statement = gens.generate_sample_model(ossp.Statement)
            statement.statement_id = control_mapping
            statement.by_components = ExportReader._add_new_by_comps(by_comps)
            implemented_req.statements = as_list(implemented_req.statements)
            implemented_req.statements.append(statement)
            implemented_req.statements = sorted(implemented_req.statements, key=lambda x: x.statement_id)
        else:
            implemented_req = self._add_or_get_implemented_requirement(control_mapping)
            implemented_req.by_components = as_list(implemented_req.by_components)
            implemented_req.by_components.extend(ExportReader._add_new_by_comps(by_comps))

    def _add_or_get_implemented_requirement(self, control_id: str) -> ossp.ImplementedRequirement:
        """Add or get implemented requirement from implemented requirements dictionary."""
        if control_id in self._implemented_requirements:
            return self._implemented_requirements[control_id]

        implemented_requirement = gens.generate_sample_model(ossp.ImplementedRequirement)
        implemented_requirement.control_id = control_id
        self._implemented_requirements[control_id] = implemented_requirement
        return implemented_requirement

    @staticmethod
    def _add_new_by_comps(by_comp_dict: ByComponentDict) -> List[ossp.ByComponent]:
        """Add new by_components to the implemented requirement."""
        new_by_comp: List[ossp.ByComponent] = []
        for comp_uuid, inheritance_info in by_comp_dict.items():
            by_comp: ossp.ByComponent = gens.generate_sample_model(ossp.ByComponent)
            by_comp.component_uuid = comp_uuid
            by_comp.inherited = none_if_empty(inheritance_info[0])
            by_comp.satisfied = none_if_empty(inheritance_info[1])
            new_by_comp.append(by_comp)
        return new_by_comp

    def _read_inheritance_markdown_directory(self) -> InheritanceViewDict:
        """Read all inheritance markdown files and return a dictionary of all the information."""
        markdown_dict: InheritanceViewDict = {}

        # Creating a dictionary to find the component uuid by title for faster lookup
        uuid_by_title: Dict[str, str] = {}
        for component in as_list(self._ssp.system_implementation.components):
            uuid_by_title[component.title] = component.uuid

        for comp_dir in os.listdir(self._root_path):
            is_comp_leveraged = False
            for control_dir in os.listdir(os.path.join(self._root_path, comp_dir)):
                control_dir_path = os.path.join(self._root_path, comp_dir, control_dir)

                # Initialize the by_component dictionary for the control directory
                # If it exists in the markdown dictionary, then update it with the new information
                by_comp_dict = markdown_dict.get(control_dir, {})

                for file in os.listdir(control_dir_path):
                    file_path = pathlib.Path(control_dir_path).joinpath(file)
                    reader = InheritanceMarkdownReader(file_path)
                    leveraged_info = reader.process_leveraged_statement_markdown()

                    # If there is no leveraged information, then skip this file
                    if leveraged_info is None:
                        continue

                    # If a file has leveraged information, then set the flag to indicate the component is leveraged
                    is_comp_leveraged = True

                    for comp in leveraged_info.leveraging_comp_titles:
                        if comp not in uuid_by_title:
                            keys_as_string = ', '.join(uuid_by_title.keys())
                            raise TrestleError(
                                f'Component "{comp}" does not exist in the {self._ssp.metadata.title} SSP. '
                                f'Please use options: {keys_as_string}.'
                            )

                        comp_uuid = uuid_by_title[comp]
                        inherited, satisfied = by_comp_dict.get(comp_uuid, ([], []))

                        if leveraged_info.inherited is not None:
                            inherited.append(leveraged_info.inherited)
                        if leveraged_info.satisfied is not None:
                            satisfied.append(leveraged_info.satisfied)

                        by_comp_dict[comp_uuid] = (inherited, satisfied)

                markdown_dict[control_dir] = by_comp_dict

            if is_comp_leveraged:
                self._mapped_components.append(comp_dir)

        return markdown_dict



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle Repository APIs."""

import os
import pathlib

import pytest

from tests import test_utils

import trestle.common.const as const
import trestle.oscal as oscal
import trestle.oscal.catalog as cat
import trestle.oscal.profile as prof
from trestle.common.err import TrestleError
from trestle.core import generators, parser
from trestle.core.repository import AgileAuthoring, ManagedOSCAL, Repository

prof_name = 'comp_prof'
ssp_name = 'my_ssp'
cat_name = 'simplified_nist_catalog'
md_dir = 'test_md'


def test_repo(tmp_trestle_dir: pathlib.Path) -> None:
    """Test creating Repository object."""
    repo = Repository(tmp_trestle_dir)
    assert repo.root_dir == tmp_trestle_dir


def test_repo_invalid_root(tmp_path: pathlib.Path) -> None:
    """Invalid trestle_root directory while creating Repository object."""
    with pytest.raises(TrestleError, match='not a valid Trestle root'):
        Repository(tmp_path)


def test_import(tmp_trestle_dir: pathlib.Path) -> None:
    """Test import."""
    # Generate sample catalog model
    catalog_data = generators.generate_sample_model(cat.Catalog)

    repo = Repository(tmp_trestle_dir)
    managed_oscal = repo.import_model(catalog_data, 'imported')
    assert managed_oscal.root_dir == tmp_trestle_dir
    assert managed_oscal.model_name == 'imported'
    assert managed_oscal.model_type == catalog_data.__class__
    assert managed_oscal.filepath.exists()


def test_import_invalid_top_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid top model."""
    # try to import Metadata
    metadata = generators.generate_sample_model(oscal.common.Metadata)

    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match='not a top level model'):
        repo.import_model(metadata, 'imported')


def test_import_model_exists(tmp_trestle_dir: pathlib.Path) -> None:
    """Model already exists."""
    # Generate sample catalog model
    catalog_data = generators.generate_sample_model(cat.Catalog)

    repo = Repository(tmp_trestle_dir)
    managed_oscal = repo.import_model(catalog_data, 'imported')
    assert managed_oscal.filepath.exists()

    with pytest.raises(TrestleError, match=r'OSCAL file .* exists'):
        repo.import_model(catalog_data, 'imported')


def test_import_validation_fail(tmp_trestle_dir: pathlib.Path) -> None:
    """Validation failed."""
    # catalog data
    dup_cat = {
        'uuid': '525f94af-8007-4376-8069-aa40179e0f6e',
        'metadata': {
            'title': 'Generic catalog created by trestle.',
            'last-modified': '2020-12-11T02:04:51.053+00:00',
            'version': '0.0.0',
            'oscal-version': oscal.OSCAL_VERSION
        },
        'back-matter': {
            'resources': [
                {
                    'uuid': 'b1101385-9e36-44a3-ba03-98b6ebe0a367'
                }, {
                    'uuid': 'b1101385-9e36-44a3-ba03-98b6ebe0a367'
                }
            ]
        }
    }
    catalog_data = parser.parse_dict(dup_cat, 'trestle.oscal.catalog.Catalog')

    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match=r'Validation .* did not pass'):
        repo.import_model(catalog_data, 'imported')


def test_list(tmp_trestle_dir: pathlib.Path) -> None:
    """Test list models."""
    # 1. Empty list
    repo = Repository(tmp_trestle_dir)
    model_list = repo.list_models(cat.Catalog)
    assert len(model_list) == 0

    # model exists
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo.import_model(catalog_data, 'imported')
    model_list = repo.list_models(cat.Catalog)
    assert len(model_list) == 1
    assert 'imported' in model_list


def test_list_invalid_top_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid top model."""
    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match='not a top level model'):
        repo.list_models(oscal.common.Metadata)


def test_get(tmp_trestle_dir: pathlib.Path) -> None:
    """Test get model."""
    # create a model
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    repo.import_model(catalog_data, 'imported')
    managed_oscal = repo.get_model(cat.Catalog, 'imported')
    assert managed_oscal.model_name == 'imported'


def test_get_invalid_top_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid top model."""
    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match='not a top level model'):
        repo.get_model(oscal.common.Metadata, 'anything')


def test_get_model_not_exists(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid get model does not exists."""
    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match='does not exist'):
        repo.get_model(cat.Catalog, 'anything')


def test_delete(tmp_trestle_dir: pathlib.Path) -> None:
    """Test delete model."""
    # create a model
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    repo.import_model(catalog_data, 'imported')
    # created model is 'dist' folder also
    repo.assemble_model(cat.Catalog, 'imported')
    success = repo.delete_model(cat.Catalog, 'imported')
    assert success


def test_delete_invalid_top_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid top model."""
    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match='not a top level model'):
        repo.delete_model(oscal.common.Metadata, 'anything')


def test_delete_model_not_exists(tmp_trestle_dir: pathlib.Path) -> None:
    """Delete model does not exists."""
    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match='does not exist'):
        repo.delete_model(cat.Catalog, 'anything')


def test_assemble(tmp_trestle_dir: pathlib.Path) -> None:
    """Test assemble model."""
    # create a model
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    repo.import_model(catalog_data, 'imported')
    success = repo.assemble_model(cat.Catalog, 'imported')
    assert success
    dist_model_path = pathlib.Path(tmp_trestle_dir, 'dist', 'catalogs', 'imported.json')
    assert dist_model_path.exists()


def test_assemble_invalid_top_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid top model."""
    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match='not a top level model'):
        repo.assemble_model(oscal.common.Metadata, 'anything')


def test_assemble_model_not_exists(tmp_trestle_dir: pathlib.Path) -> None:
    """Assemble model does not exists."""
    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError):
        repo.assemble_model(cat.Catalog, 'anything')


def test_validate(tmp_trestle_dir: pathlib.Path) -> None:
    """Test validate model."""
    # create a model
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    repo.import_model(catalog_data, 'imported')
    success = repo.validate_model(cat.Catalog, 'imported')
    assert success


def test_validate_invalid_top_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid top model."""
    repo = Repository(tmp_trestle_dir)
    with pytest.raises(TrestleError, match='not a top level model'):
        repo.validate_model(oscal.common.Metadata, 'anything')


def test_validate_model_not_exists(tmp_trestle_dir: pathlib.Path) -> None:
    """Assemble model does not exists."""
    repo = Repository(tmp_trestle_dir)
    success = repo.validate_model(cat.Catalog, 'anything')
    assert not success


def test_managed_oscal(tmp_trestle_dir: pathlib.Path) -> None:
    """Test creating Managed OSCAL object."""
    # generate catalog data and import
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')
    assert managed.model_dir == tmp_trestle_dir / 'catalogs' / 'imported'


def test_managed_invalid_root(tmp_path: pathlib.Path) -> None:
    """Invalid trestle_root directory while creating Managed OSCAL object."""
    with pytest.raises(TrestleError, match='not a valid Trestle root'):
        ManagedOSCAL(tmp_path, cat.Catalog, 'anything')


def test_managed_invalid_top_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid top model while creating Managed OSCAL object."""
    with pytest.raises(TrestleError, match='not a top level model'):
        ManagedOSCAL(tmp_trestle_dir, oscal.catalog.Group, 'anything')


def test_managed_invalid_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid model directory while creating Managed OSCAL object."""
    with pytest.raises(TrestleError, match=r'Model .* does not exist'):
        ManagedOSCAL(tmp_trestle_dir, cat.Catalog, 'anything')


def test_managed_file_not_exist(tmp_trestle_dir: pathlib.Path) -> None:
    """Test model file does not exist while creating a Managed OSCAL object."""
    # generate catalog data and import
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')

    # delete file
    managed.filepath.unlink()
    with pytest.raises(TrestleError, match=r'Model file .* does not exist'):
        ManagedOSCAL(tmp_trestle_dir, cat.Catalog, 'imported')


def test_managed_read(tmp_trestle_dir: pathlib.Path) -> None:
    """Test model read."""
    # generate catalog data and import
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')
    model = managed.read()
    assert model.uuid == catalog_data.uuid


def test_managed_write(tmp_trestle_dir: pathlib.Path) -> None:
    """Test model write."""
    # generate catalog data and import
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')

    # generate another catalog data for writing
    catalog_data = generators.generate_sample_model(cat.Catalog)
    success = managed.write(catalog_data)
    assert success


def test_managed_write_invalid_top_model(tmp_trestle_dir: pathlib.Path) -> None:
    """Invalid top level model while writing."""
    # generate catalog data and import
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')

    # generate another catalog data for writing
    catalog_data = generators.generate_sample_model(oscal.catalog.Group)

    with pytest.raises(TrestleError, match='not a top level model'):
        managed.write(catalog_data)


def test_managed_split(tmp_trestle_dir: pathlib.Path) -> None:
    """Test model split."""
    # generate catalog data and import
    filepath = test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME
    catalog_data = cat.Catalog.oscal_read(filepath)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')

    # store current working directory
    cwd = pathlib.Path.cwd()

    # test splitting
    success = managed.split(pathlib.Path('catalog.json'), ['catalog.metadata'])
    assert success
    assert pathlib.Path(tmp_trestle_dir / 'catalogs' / 'imported' / 'catalog' / 'metadata.json').exists()

    # test cwd is restored after splitting
    assert pathlib.Path.cwd() == cwd

    success = managed.split(pathlib.Path('catalog/metadata.json'), ['metadata.props'])
    assert success
    assert pathlib.Path(tmp_trestle_dir / 'catalogs' / 'imported' / 'catalog' / 'metadata' / 'props.json').exists()


def test_managed_split_multi(tmp_trestle_dir: pathlib.Path) -> None:
    """Test model split multiple elements."""
    # generate catalog data and import
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')

    # store current working directory
    cwd = pathlib.Path.cwd()

    # test splitting
    success = managed.split(pathlib.Path('catalog.json'), ['catalog.metadata', 'catalog.groups.*'])
    assert success

    # test cwd is restored after splitting
    assert pathlib.Path.cwd() == cwd


def test_managed_merge(tmp_trestle_dir: pathlib.Path) -> None:
    """Test model merge."""
    # generate catalog data and import and split
    filepath = test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME
    catalog_data = cat.Catalog.oscal_read(filepath)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')

    # split should be success
    success = managed.split(pathlib.Path('catalog.json'), ['catalog.metadata'])
    assert success
    assert pathlib.Path(tmp_trestle_dir / 'catalogs' / 'imported' / 'catalog' / 'metadata.json').exists()

    success = managed.split(pathlib.Path('catalog/metadata.json'), ['metadata.props'])
    assert success
    assert pathlib.Path(tmp_trestle_dir / 'catalogs' / 'imported' / 'catalog' / 'metadata' / 'props.json').exists()

    # store current working directory before merge
    cwd = pathlib.Path.cwd()

    # merge should be success
    success = managed.merge(['metadata.*'], pathlib.Path('catalog'))
    assert success
    assert not pathlib.Path(tmp_trestle_dir / 'catalogs' / 'imported' / 'catalog' / 'metadata' / 'props.json').exists()

    success = managed.merge(['catalog.*'])
    assert success
    assert not pathlib.Path(tmp_trestle_dir / 'catalogs' / 'imported' / 'catalog' / 'metadata.json').exists()

    # test cwd is restored after splitting
    assert pathlib.Path.cwd() == cwd


def test_managed_validate(tmp_trestle_dir: pathlib.Path) -> None:
    """Test model validate."""
    # generate catalog data and import
    catalog_data = generators.generate_sample_model(cat.Catalog)
    repo = Repository(tmp_trestle_dir)
    managed = repo.import_model(catalog_data, 'imported')
    success = managed.validate()
    assert success


# Test default agile authoring paths to ensure call from repository are correct


def test_agile_authoring_catalog(tmp_trestle_dir: pathlib.Path) -> None:
    """Test catalog generate and assemble through API."""
    test_utils.load_from_json(tmp_trestle_dir, cat_name, cat_name, cat.Catalog)

    authoring = AgileAuthoring(tmp_trestle_dir)

    md_cat = os.path.join(md_dir, cat_name)
    success = authoring.generate_catalog_markdown(cat_name, md_cat)

    assert success
    assert pathlib.Path(tmp_trestle_dir / md_cat).exists()

    new_cat = 'temp_cat'
    success = authoring.assemble_catalog_markdown(cat_name, new_cat, md_cat)
    assert success
    assert pathlib.Path(tmp_trestle_dir, const.MODEL_DIR_CATALOG, new_cat).exists()


def test_agile_authoring_profile(tmp_trestle_dir: pathlib.Path) -> None:
    """Test profile generate and assemble through API."""
    test_utils.load_from_json(tmp_trestle_dir, cat_name, cat_name, cat.Catalog)
    test_utils.load_from_json(tmp_trestle_dir, prof_name, prof_name, prof.Profile)

    authoring = AgileAuthoring(tmp_trestle_dir)

    md_prof = os.path.join(md_dir, prof_name)
    success = authoring.generate_profile_markdown(prof_name, md_prof)

    assert success
    assert pathlib.Path(tmp_trestle_dir / md_prof).exists()

    new_prof = 'temp_prof'
    success = authoring.assemble_profile_markdown(prof_name, new_prof, md_prof)
    assert success
    assert pathlib.Path(tmp_trestle_dir, const.MODEL_DIR_PROFILE, new_prof).exists()


def test_agile_authoring_component(tmp_trestle_dir: pathlib.Path) -> None:
    """Test component generate and assemble through API."""
    comp_name = test_utils.setup_component_generate(tmp_trestle_dir)
    authoring = AgileAuthoring(tmp_trestle_dir)

    md_comp = os.path.join(md_dir, comp_name)
    success = authoring.generate_component_definition_markdown(comp_name, md_comp)

    assert success
    assert pathlib.Path(tmp_trestle_dir / md_comp).exists()

    new_comp = 'temp_comp'
    success = authoring.assemble_component_definition_markdown(comp_name, new_comp, md_comp)
    assert success
    assert pathlib.Path(tmp_trestle_dir, const.MODEL_DIR_COMPDEF, new_comp).exists()


def test_agile_authoring_ssp(tmp_trestle_dir: pathlib.Path) -> None:
    """Test ssp generate and assemble through API."""
    args, _ = test_utils.setup_for_ssp(tmp_trestle_dir, prof_name, ssp_name)
    authoring = AgileAuthoring(tmp_trestle_dir)

    success = authoring.generate_ssp_markdown(args.profile, args.output, args.compdefs)

    assert success
    assert pathlib.Path(tmp_trestle_dir / args.output).exists()

    success = authoring.assemble_ssp_markdown(ssp_name, ssp_name, args.output, args.compdefs)
    assert success
    assert pathlib.Path(tmp_trestle_dir, const.MODEL_DIR_SSP, args.output).exists()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script to generate python models from oscal using datamodel-code-generator."""
import logging
import re
import sys
from pathlib import Path
from subprocess import CalledProcessError, check_call

from oscal_normalize import normalize_files

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


def load_git():
    """Load git submodule for oscal."""
    # NOTE: this should only be done if the latest nist content is desired
    # otherwise trestle 1.0.0 should track the corresponding submodule it is bound to
    # nist source checkout is bound to tags/v1.0.0 for trestle 1.0.0
    logger.info('git add and update oscal modules')
    try:
        check_call('git submodule add https://github.com/usnistgov/OSCAL.git nist-source'.split())
    except CalledProcessError:
        # silently ignore already existing module
        pass
    # Add second module
    try:
        check_call('git submodule add https://github.com/usnistgov/oscal-content.git nist-content'.split())
    except CalledProcessError:
        # silently ignore already existing module
        pass
    try:
        check_call('git submodule update --init'.split())
    except CalledProcessError as error:
        logger.error(f'Error updating the oscal git submodule {error}')
    try:
        check_call('git submodule update --remote --merge'.split())
    except CalledProcessError as error:
        logger.error(f'Error updating the oscal git submodule {error}')


def generate_model(full_name, out_full_name):
    """Generate a single model with datamodel-codegen."""
    logger.info(f'generate python model with datamodel-codegen: {str(full_name)} -> {str(out_full_name)}')
    args = [
        'datamodel-codegen',
        '--disable-timestamp',
        '--disable-appending-item-suffix',
        '--use-schema-description',
        '--input-file-type',
        'jsonschema',
        '--input',
        str(full_name),
        '--base-class',
        'trestle.core.base_model.OscalBaseModel',
        '--output',
        str(out_full_name)
    ]
    try:
        check_call(args)
    except CalledProcessError as error:
        logger.error(f'Error calling datamodel-codegen for file {full_name} error {error}')


def generate_models():
    """
    Generate all models including 3rd party.

    IMPORTANT NOTE: This script will leave the temporary classes in oscal/tmp for reference, but they should be deleted
    when no longer needed because some of the operations in the build will seek all .py files, and will try to act
    on these temporary files - giving an error.  So be sure to delete the oscal/tmp directory when done generating
    new classes.
    """
    logger.info('generating models')
    out_dir = Path('trestle/oscal')
    out_dir.mkdir(exist_ok=True, parents=True)
    tmp_dir = out_dir / 'tmp'
    tmp_dir.mkdir(exist_ok=True, parents=True)
    out_init = out_dir / '__init__.py'
    out_init.touch(exist_ok=True)

    in_dir = Path('nist-source/json/schema')
    for full_name in in_dir.glob('oscal_*_schema.json'):
        if 'complete' in str(full_name):
            continue
        try:
            obj = re.search('oscal_(.+?)_schema.json', str(full_name)).group(1)
        except AttributeError:
            logger.error(f'Warning: filename did not parse properly: {full_name}')
            obj = None
            continue
        oscal_name = obj.replace('-', '_')
        out_fname = oscal_name + '.py'
        out_full_name = tmp_dir / out_fname
        generate_model(full_name, out_full_name)
    # all .py files are first generated into oscal/tmp to be normalized'
    normalize_files()


def main():
    """Load git and generate models."""
    # at this point should load_git() if latest oscal schemas are needed
    generate_models()
    logger.info('DONE')


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the ssp_generator module."""

import pathlib
import uuid

import pytest

from tests import test_utils

import trestle.common.const as const
import trestle.core.crm.export_reader as exportreader
import trestle.core.generators as gens
import trestle.oscal.ssp as ossp
from trestle.common.err import TrestleError
from trestle.common.model_utils import ModelUtils
from trestle.core.models.file_content_type import FileContentType

leveraged_ssp = 'leveraged_ssp'
leveraging_ssp = 'my_ssp'

expected_appliance_uuid = '22222222-0000-4000-9001-000000000003'
expected_saas_uuid = '22222222-0000-4000-9001-000000000001'

example_provided_uuid = '18ac4e2a-b5f2-46e4-94fa-cc84ab6fe114'
example_responsibility_uuid = '4b34c68f-75fa-4b38-baf0-e50158c13ac2'


@pytest.fixture(scope='function')
def sample_implemented_requirement() -> ossp.ImplementedRequirement:
    """Return a valid ComponentDefinition object with some contents."""
    # one component has no properties - the other has two
    impl_req: ossp.ImplementedRequirement = gens.generate_sample_model(ossp.ImplementedRequirement)
    by_comp: ossp.ByComponent = gens.generate_sample_model(ossp.ByComponent)
    impl_req.by_components = [by_comp]
    return impl_req


def prep_inheritance_dir(ac_appliance_dir: pathlib.Path, inheritance_text: str) -> None:
    """Prepare inheritance directory with basic information."""
    ac_2 = ac_appliance_dir.joinpath('ac-2')
    ac_2.mkdir(parents=True)

    file = ac_2 / f'{expected_appliance_uuid}.md'
    with open(file, 'w') as f:
        f.write(inheritance_text)

    # test with a statement
    ac_2a = ac_appliance_dir.joinpath('ac-2_smt.a')
    ac_2a.mkdir(parents=True)

    file = ac_2a / f'{expected_appliance_uuid}.md'
    with open(file, 'w') as f:
        f.write(inheritance_text)


def test_read_exports_from_markdown(tmp_trestle_dir: pathlib.Path) -> None:
    """Test exports reader with inheritance view."""
    inheritance_path = tmp_trestle_dir.joinpath(leveraged_ssp, const.INHERITANCE_VIEW_DIR)
    ac_appliance_dir = inheritance_path.joinpath('Access Control Appliance')
    inheritance_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=['Access Control Appliance', 'THIS SYSTEM (SaaS)'],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )
    prep_inheritance_dir(ac_appliance_dir, inheritance_text)

    test_utils.load_from_json(tmp_trestle_dir, 'leveraging_ssp', leveraging_ssp, ossp.SystemSecurityPlan)

    orig_ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        leveraging_ssp,
        ossp.SystemSecurityPlan,
        FileContentType.JSON)

    reader = exportreader.ExportReader(inheritance_path, orig_ssp)
    ssp = reader.read_exports_from_markdown()

    implemented_requirements = ssp.control_implementation.implemented_requirements

    assert implemented_requirements[0].control_id == 'ac-2'
    assert implemented_requirements[0].by_components[0].component_uuid == expected_appliance_uuid  # type: ignore

    by_comp = implemented_requirements[0].by_components[0]  # type: ignore

    assert by_comp.inherited[0].provided_uuid == '18ac4e2a-b5f2-46e4-94fa-cc84ab6fe114'  # type: ignore
    assert by_comp.satisfied[0].responsibility_uuid == '4b34c68f-75fa-4b38-baf0-e50158c13ac2'  # type: ignore
    assert by_comp.satisfied[0].description == 'My Satisfied Description'  # type: ignore

    assert implemented_requirements[0].by_components[1].component_uuid == expected_saas_uuid  # type: ignore
    by_comp = implemented_requirements[0].by_components[1]  # type: ignore

    assert by_comp.inherited[0].provided_uuid == '18ac4e2a-b5f2-46e4-94fa-cc84ab6fe114'  # type: ignore
    assert by_comp.satisfied[0].responsibility_uuid == '4b34c68f-75fa-4b38-baf0-e50158c13ac2'  # type: ignore
    assert by_comp.satisfied[0].description == 'My Satisfied Description'  # type: ignore

    # Ensure that the statement is also added to the SSP
    assert implemented_requirements[0].statements is not None
    assert implemented_requirements[0].statements[0].statement_id == 'ac-2_smt.a'


def test_read_inheritance_markdown_dir(tmp_trestle_dir: pathlib.Path) -> None:
    """Test reading inheritance view directory."""
    inheritance_path = tmp_trestle_dir.joinpath(leveraged_ssp, const.INHERITANCE_VIEW_DIR)
    ac_appliance_dir = inheritance_path.joinpath('Access Control Appliance')
    inheritance_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=['Access Control Appliance', 'THIS SYSTEM (SaaS)'],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )
    prep_inheritance_dir(ac_appliance_dir, inheritance_text)

    unmapped_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=[const.REPLACE_ME],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )

    ac_21 = ac_appliance_dir.joinpath('ac-2.1')
    ac_21.mkdir(parents=True)
    # Ensure this file does not get added to the dictionary
    file = ac_21 / f'{expected_appliance_uuid}.md'
    with open(file, 'w') as f:
        f.write(unmapped_text)

    test_utils.load_from_json(tmp_trestle_dir, 'leveraging_ssp', leveraging_ssp, ossp.SystemSecurityPlan)

    orig_ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        leveraging_ssp,
        ossp.SystemSecurityPlan,
        FileContentType.JSON)

    reader = exportreader.ExportReader(inheritance_path, orig_ssp)
    markdown_dict: exportreader.InheritanceViewDict = reader._read_inheritance_markdown_directory()

    assert len(markdown_dict) == 3
    assert 'ac-2' in markdown_dict
    assert len(markdown_dict['ac-2']) == 2
    assert expected_appliance_uuid in markdown_dict['ac-2']

    assert len(markdown_dict['ac-2.1']) == 0

    inheritance_info = markdown_dict['ac-2'][expected_appliance_uuid]

    assert inheritance_info[0][0].provided_uuid == '18ac4e2a-b5f2-46e4-94fa-cc84ab6fe114'
    assert inheritance_info[1][0].responsibility_uuid == '4b34c68f-75fa-4b38-baf0-e50158c13ac2'
    assert inheritance_info[1][0].description == 'My Satisfied Description'


def test_read_inheritance_markdown_dir_with_multiple_leveraged_components(tmp_trestle_dir: pathlib.Path) -> None:
    """Test reading inheritance view directory with components that span multiple leveraged components."""
    inheritance_path = tmp_trestle_dir.joinpath(leveraged_ssp, const.INHERITANCE_VIEW_DIR)

    ac_appliance_dir = inheritance_path.joinpath('Access Control Appliance')
    inheritance_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=['Access Control Appliance', 'THIS SYSTEM (SaaS)'],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )
    prep_inheritance_dir(ac_appliance_dir, inheritance_text)

    inheritance_text_2 = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=['Access Control Appliance'],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )

    unmapped_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=[const.REPLACE_ME],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )

    this_system_dir = inheritance_path.joinpath('This System')
    ac_2 = this_system_dir.joinpath('ac-2')
    ac_2.mkdir(parents=True)

    file = ac_2 / f'{expected_appliance_uuid}.md'
    with open(file, 'w') as f:
        f.write(inheritance_text_2)

    ac_2a = this_system_dir.joinpath('ac-2_smt.a')
    ac_2a.mkdir(parents=True)

    file = ac_2a / f'{expected_appliance_uuid}.md'
    with open(file, 'w') as f:
        f.write(unmapped_text)

    test_utils.load_from_json(tmp_trestle_dir, 'leveraging_ssp', leveraging_ssp, ossp.SystemSecurityPlan)

    orig_ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        leveraging_ssp,
        ossp.SystemSecurityPlan,
        FileContentType.JSON)

    reader = exportreader.ExportReader(inheritance_path, orig_ssp)
    markdown_dict: exportreader.InheritanceViewDict = reader._read_inheritance_markdown_directory()

    assert len(markdown_dict) == 2
    assert 'ac-2' in markdown_dict
    assert len(markdown_dict['ac-2']) == 2

    assert expected_appliance_uuid in markdown_dict['ac-2']
    inheritance_info = markdown_dict['ac-2'][expected_appliance_uuid]

    assert len(inheritance_info[0]) == 2
    assert len(inheritance_info[1]) == 2

    assert 'ac-2_smt.a' in markdown_dict
    assert len(markdown_dict['ac-2_smt.a']) == 2

    assert expected_appliance_uuid in markdown_dict['ac-2_smt.a']
    inheritance_info = markdown_dict['ac-2_smt.a'][expected_appliance_uuid]

    # Only leveraging from one component
    assert len(inheritance_info[0]) == 1
    assert len(inheritance_info[1]) == 1


def test_read_inheritance_markdown_dir_with_invalid_mapping(tmp_trestle_dir: pathlib.Path) -> None:
    """Test reading inheritance view directory with a component that does not exist."""
    inheritance_path = tmp_trestle_dir.joinpath(leveraged_ssp, const.INHERITANCE_VIEW_DIR)

    invalid_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=['Invalid Component'],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )

    this_system_dir = inheritance_path.joinpath('This System')
    ac_2 = this_system_dir.joinpath('ac-2')
    ac_2.mkdir(parents=True)

    file = ac_2 / f'{expected_appliance_uuid}.md'
    with open(file, 'w') as f:
        f.write(invalid_text)

    test_utils.load_from_json(tmp_trestle_dir, 'leveraging_ssp', leveraging_ssp, ossp.SystemSecurityPlan)

    orig_ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        leveraging_ssp,
        ossp.SystemSecurityPlan,
        FileContentType.JSON)

    with pytest.raises(TrestleError):
        reader = exportreader.ExportReader(inheritance_path, orig_ssp)
        _ = reader._read_inheritance_markdown_directory()


def test_get_leveraged_ssp_reference(tmp_trestle_dir: pathlib.Path) -> None:
    """Test retrieving leveraged SSP reference from Markdown."""
    inheritance_path = tmp_trestle_dir.joinpath(leveraged_ssp, const.INHERITANCE_VIEW_DIR)

    ac_appliance_dir = inheritance_path.joinpath('Access Control Appliance')
    inheritance_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=['Access Control Appliance', 'THIS SYSTEM (SaaS)'],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )
    prep_inheritance_dir(ac_appliance_dir, inheritance_text)

    test_utils.load_from_json(tmp_trestle_dir, 'leveraging_ssp', leveraging_ssp, ossp.SystemSecurityPlan)

    orig_ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        leveraging_ssp,
        ossp.SystemSecurityPlan,
        FileContentType.JSON)

    reader = exportreader.ExportReader(inheritance_path, orig_ssp)
    assert reader.get_leveraged_ssp_href() == 'trestle://leveraged_ssp.json'


def test_get_leveraged_components(tmp_trestle_dir: pathlib.Path) -> None:
    """Test leveraged mapped components from Markdown."""
    inheritance_path = tmp_trestle_dir.joinpath(leveraged_ssp, const.INHERITANCE_VIEW_DIR)

    ac_appliance_dir = inheritance_path.joinpath('Access Control Appliance')
    unmapped_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=[const.REPLACE_ME],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )
    prep_inheritance_dir(ac_appliance_dir, unmapped_text)

    this_system_dir = inheritance_path.joinpath('This System')
    ac_2 = this_system_dir.joinpath('ac-2')
    ac_2.mkdir(parents=True)

    inheritance_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=['Access Control Appliance', 'THIS SYSTEM (SaaS)'],
        leveraged_ssp_href='trestle://leveraged_ssp.json'
    )

    file = ac_2 / f'{expected_appliance_uuid}.md'
    with open(file, 'w') as f:
        f.write(inheritance_text)

    test_utils.load_from_json(tmp_trestle_dir, 'leveraging_ssp', leveraging_ssp, ossp.SystemSecurityPlan)

    orig_ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        leveraging_ssp,
        ossp.SystemSecurityPlan,
        FileContentType.JSON)

    reader = exportreader.ExportReader(inheritance_path, orig_ssp)
    _ = reader.read_exports_from_markdown()

    leveraged_components = reader.get_leveraged_components()

    assert len(leveraged_components) == 1
    assert 'Access Control Appliance' not in leveraged_components
    assert 'This System' in leveraged_components


def test_update_type_with_by_comp(sample_implemented_requirement: ossp.ImplementedRequirement) -> None:
    """Test update type with by component."""
    test_ssp: ossp.SystemSecurityPlan = gens.generate_sample_model(ossp.SystemSecurityPlan)
    reader = exportreader.ExportReader('', test_ssp)

    test_inherited: ossp.Inherited = gens.generate_sample_model(ossp.Inherited)
    test_satisfied: ossp.Satisfied = gens.generate_sample_model(ossp.Satisfied)

    test_comp_uuid = str(uuid.uuid4())

    test_by_comp_dict: exportreader.ByComponentDict = {test_comp_uuid: ([test_inherited], [test_satisfied])}

    assert len(sample_implemented_requirement.by_components) == 1

    reader._update_type_with_by_comp(sample_implemented_requirement, test_by_comp_dict)

    # Ensure a new by_comp was added, but the original was not removed
    assert len(sample_implemented_requirement.by_components) == 2

    # Test update the existing without adding a new component
    test_satisfied.description = 'Updated Description'
    test_by_comp_dict: exportreader.ByComponentDict = {test_comp_uuid: ([test_inherited], [test_satisfied])}
    reader._update_type_with_by_comp(sample_implemented_requirement, test_by_comp_dict)

    assert len(sample_implemented_requirement.by_components) == 2
    new_by_comp = sample_implemented_requirement.by_components[1]  # type: ignore

    assert new_by_comp.component_uuid == test_comp_uuid
    assert new_by_comp.satisfied is not None
    assert new_by_comp.satisfied[0].description == 'Updated Description'

    # Test removing the existing inheritance info
    test_by_comp_dict: exportreader.ByComponentDict = {}
    reader._update_type_with_by_comp(sample_implemented_requirement, test_by_comp_dict)

    new_by_comp = sample_implemented_requirement.by_components[1]  # type: ignore

    assert new_by_comp.component_uuid == test_comp_uuid
    assert new_by_comp.satisfied is None
    assert new_by_comp.inherited is None



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Element wrapper of an OSCAL model element."""

import logging
import pathlib
from typing import Any, List, Optional, Type, Union, cast

from pydantic import Field, create_model
from pydantic.error_wrappers import ValidationError

from ruamel.yaml import YAML

import trestle.common.const as const
from trestle.common import common_types, str_utils, type_utils as utils
from trestle.common.err import TrestleError, TrestleNotFoundError
from trestle.common.model_utils import ModelUtils
from trestle.common.str_utils import AliasMode, classname_to_alias
from trestle.core.base_model import OscalBaseModel
from trestle.core.models.file_content_type import FileContentType

logger = logging.getLogger(__name__)


class ElementPath:
    """Element path wrapper of an element.

    This only allows a single wildcard '*' at the end to denote elements of an array or dict
    """

    PATH_SEPARATOR: str = const.ALIAS_PATH_SEPARATOR

    WILDCARD: str = '*'

    def __init__(self, element_path: str, parent_path: Optional['ElementPath'] = None) -> None:
        """Initialize an element wrapper.

        It assumes the element path contains oscal field alias with hyphens only
        """
        self._parent_path = parent_path

        self._path: List[str] = self._parse(element_path)

        # Initialize private variables for lazy processing and caching
        self._element_name: Optional[str] = None
        self._preceding_path: Optional['ElementPath'] = None

    def _parse(self, element_path: str) -> List[str]:
        """Parse the element path and validate."""
        parts: List[str] = element_path.split(self.PATH_SEPARATOR)

        for part in parts:
            if part == '':
                raise TrestleError(
                    f'Invalid path "{element_path}" because there are empty path parts between "{self.PATH_SEPARATOR}" '
                    'or in the beginning'
                )

        if parts[0] == self.WILDCARD:
            raise TrestleError(f'Invalid path {element_path} with wildcard.')
        return parts

    def get(self) -> List[str]:
        """Return the path parts as a list."""
        return self._path

    def get_type(self, root_model: Optional[Type[Any]] = None, use_parent: bool = False) -> Type[Any]:
        """Get the type of an element.

        If possible the model type will be derived from one of the top level models,
        otherwise a 'root model' can be passed for situations where this is not possible.

        This type path should *NOT* have wild cards in it. It *may* have* indices.
        Valid Examples:
            catalog.metadata
            catalog.groups
            catalog.groups.group
            catalog
            catalog.groups.0

        Args:
            root_model: An OscalBaseModel Type from which to base the approach on.
            use_parent: Whether or not to normalise the full path across parent ElementPaths, default to not.

        Returns:
            The type of the model whether or not it is an OscalBaseModel or not.
        """
        effective_path: List[str]
        if use_parent:
            effective_path = self.get_full_path_parts()
        else:
            effective_path = self._path

        if not root_model:
            # lookup root model from top level oscal models or fail
            prev_model = self._top_level_type_lookup(effective_path[0])
        else:
            prev_model = root_model
        if len(effective_path) == 1:
            return prev_model
        # variables
        # for current_element_str in effective_path[1:]:
        for current_element_str in effective_path[1:]:
            # Determine if the parent model is a collection.
            if utils.is_collection_field_type(prev_model):
                inner_model = utils.get_inner_type(prev_model)
                inner_class_name = classname_to_alias(inner_model.__name__, AliasMode.JSON)
                # Assert that the current name fits an expected form.
                # Valid choices here are *, integer (for arrays) and the inner model alias
                if (inner_class_name == current_element_str or current_element_str == self.WILDCARD
                        or current_element_str.isnumeric()):
                    prev_model = inner_model

                else:
                    raise TrestleError('Unexpected key in element path when finding type.')

            else:
                # Indices, * are not allowed on non-collection types
                if current_element_str == self.WILDCARD:
                    raise TrestleError(
                        'Wild card in unexpected position when trying to find class type.'
                        + ' Element path type lookup can only occur where a single type can be identified.'
                    )
                prev_model = prev_model.alias_to_field_map()[current_element_str].outer_type_
        return prev_model

    def get_obm_wrapped_type(self,
                             root_model: Optional[Type[Any]] = None,
                             use_parent: bool = False) -> Type[OscalBaseModel]:
        """Get the type of the element. Wraps the collection type in an OscalBaseModel as a __root__ element.

        This should principally be used for validating content.

        Args:
            root_model: An OscalBaseModel Type from which to base the approach on.
            use_parent: Whether or not to normalise the full path across parent ElementPaths, default to not.

        Returns:
            The type of the model whether wrapped or not as an OscalBaseModel.
        """
        base_type = self.get_type(root_model, use_parent)
        # Get an outer model type.
        origin_type = utils.get_origin(base_type)

        if origin_type in [list, dict]:
            # OSCAL does not support collections of collections directly. We should not hit this scenario
            collection_name = self.get_last()
            if collection_name == self.WILDCARD:
                logger.critical('Unexpected error in type system when inferring type from element path.')
                logger.critical('Please report this issue.')
                raise TrestleError('Unknown error inferring type from element path.')
            # Final path must be the alias

            new_base_type = create_model(
                str_utils.alias_to_classname(collection_name, AliasMode.JSON),
                __base__=OscalBaseModel,
                __root__=(base_type, ...)
            )
            return new_base_type
        return base_type

    def _top_level_type_lookup(self, element_str: str) -> Type[common_types.TopLevelOscalModel]:
        """From an individual element tag, induce the type of the model.

        Args:
            element_str: individual element as text such as 'catalog' or 'profile'

        Returns:
            Top level object model such as catalog, profile etc.
        """
        # Even though awkward use chain of models.
        if element_str not in const.MODEL_TYPE_LIST:
            raise TrestleError(f'{element_str} is not a top level model (e.g. catalog, profile)')
        model_package = const.MODEL_TYPE_TO_MODEL_MODULE[element_str]
        object_type, _ = ModelUtils.get_root_model(model_package)
        object_type = cast(Type[common_types.TopLevelOscalModel], object_type)
        return object_type

    def is_multipart(self) -> bool:
        """Assert whether or not an element path is multiple parts.

        Originally element paths had to have multiple paths.
        This provides a check for higher level code that still has that requirement.

        Single part:
            catalog
            control
            assessment-results

        Multipart:
            catalog.metadata
            catalog.controls.control
        """
        return len(self._path) > 1

    def to_string(self) -> str:
        """Return the path parts as a dot-separated string."""
        return self.PATH_SEPARATOR.join(self.get())

    def get_parent(self) -> 'ElementPath':
        """Return the parent path.

        It can be None or a valid ElementPath
        """
        return self._parent_path

    def get_first(self) -> str:
        """Return the first part of the path."""
        return self._path[0]

    def get_last(self) -> str:
        """Return the last part of the path."""
        return self._path[-1]

    def get_full(self) -> str:
        """Return the full path including parent path parts as a dot separated str."""
        all_parts = self.get_full_path_parts()
        return self.PATH_SEPARATOR.join(all_parts)

    def get_element_name(self) -> str:
        """Return the element alias name from the path.

        Essentailly this the last part of the element path
        """
        # if it is available then return otherwise compute
        if self._element_name is None:
            element_name = self.get_last()
            if element_name == self.WILDCARD:
                element_name = self._path[-2]

            self._element_name = element_name

        return self._element_name

    def get_full_path_parts(self) -> List[str]:
        """Get full path parts to the element including parent path parts as a list."""
        path_parts = []
        if self.get_parent() is not None:
            parent_path_parts = self.get_parent().get_full_path_parts()
            path_parts.extend(parent_path_parts)
            path_parts.extend(self.get()[1:])  # don't use the first part
        else:
            path_parts.extend(self.get())

        return path_parts

    def get_preceding_path(self) -> 'ElementPath':
        """Return the element path to the preceding element in the path."""
        # if it is available then return otherwise compute
        if self._preceding_path is None:
            path_parts = self.get_full_path_parts()

            if len(path_parts) > 1:
                prec_path_parts = path_parts[:-1]
                self._preceding_path = ElementPath(self.PATH_SEPARATOR.join(prec_path_parts))

        return self._preceding_path

    def find_last_file_in_path(self, content_type: FileContentType, model_dir: pathlib.Path) -> pathlib.Path:
        """Find the last (nearest) existing file in the element path leading to this element."""
        # model dir is the top level dir for this model, e.g. catalogs/mycat
        path = model_dir
        extension = FileContentType.to_file_extension(content_type)
        good_model: pathlib.Path = None
        for element in self._path:
            if element == '*':
                break
            model_file = (path / element).with_suffix(extension)
            if not model_file.exists():
                break
            path = path / element
            good_model = model_file
        return good_model

    def make_absolute(self, model_dir: pathlib.Path, reference_dir: pathlib.Path):
        """Make the parts absolute from the top model dir."""
        # Match the current relative element path to the model directory and reference directory
        # If the element path is partial and doesn't connect to the top of the model,
        # need to deduce absolute element path from the model_dir and the reference directory
        # that corresponds to the root of the element path

        # if first element is a model type it is already absolute
        if self._path[0] not in const.MODEL_TYPE_LIST:
            rel_path = list(reference_dir.relative_to(model_dir).parts)
            rel_path.extend(self._path)
            self._path = rel_path

    def make_relative(self, model_relative_path: pathlib.Path) -> int:
        """Make the parts relative to the model path."""
        # The element path should currently be absolute
        # The model relative path should be relative to the top leve of the model
        # Change the element path to be relative to the model being loaded
        # Returns 0 on success and 1 on failur
        rel_path_parts = model_relative_path.parts[:-1]
        n_rel_parts = len(rel_path_parts)
        # the element path can't start above the model path
        if n_rel_parts >= len(self._path):
            return 1
        # confirm the leading parts match
        for ii in range(n_rel_parts):
            if rel_path_parts[ii] != self._path[ii]:
                return 1
        # chop off the leading parts of the absolute element path
        self._path = self._path[n_rel_parts:]
        return 0

    def to_file_path(self, content_type: FileContentType = None, root_dir: str = '') -> pathlib.Path:
        """Convert to a file or directory path for the element path.

        if content_type is not passed, it will return a path for directory
        """
        path_parts = self.get()

        # skip wildcard
        if path_parts[-1] == ElementPath.WILDCARD:
            path_parts = path_parts[:-1]

        if root_dir != '':
            path_parts[0] = root_dir

        path_str = '/'.join(path_parts)

        # add file extension if required
        # this will be omitted if it is a dir path
        if content_type is not None:
            file_extension = FileContentType.to_file_extension(content_type)
            path_str = path_str + file_extension

        # prepare the path
        file_path: pathlib.Path = pathlib.Path(f'./{path_str}')

        return file_path

    def to_root_path(self, content_type: FileContentType = None) -> pathlib.Path:
        """Convert to a file path for the element root."""
        path_str = f'./{self.get_first()}'
        if content_type is not None:
            file_extension = FileContentType.to_file_extension(content_type)
            path_str = path_str + file_extension

        file_path: pathlib.Path = pathlib.Path(path_str)
        return file_path

    def __str__(self) -> str:
        """Return string representation of element path."""
        return self.to_string()

    def __eq__(self, other) -> bool:
        """Override equality method."""
        if not isinstance(other, ElementPath):
            return False

        return self.get() == other.get()


class Element:
    """Element wrapper of an OSCAL model."""

    IGNORE_WRAPPER_ALIAS = '__'

    _allowed_sub_element_types: List[str] = ['Element', 'OscalBaseModel', 'list', 'None', 'dict']

    def __init__(self, elem: OscalBaseModel, wrapper_alias: str = ''):
        """Initialize an element wrapper.

        wrapper_alias is the OSCAL alias for the given elem object and used for seriazation in to_json() method.

        For example,
         - List[Catalog.Group] element should have wrapper alias 'groups'
         - Catalog element should have wrapper alias 'catalog'

        wrapper_alias is deduced for collection type object

        if wrapper_alias = IGNORE_WRAPPER_ALIAS, then it is ignored and assumed to be json-serializable during to_json()
        """
        # FIXME: There are instances where elem is a list.
        self._elem: OscalBaseModel = elem

        if wrapper_alias == '' and wrapper_alias != self.IGNORE_WRAPPER_ALIAS:
            class_name = elem.__class__.__name__
            if utils.is_collection_field_type(elem):
                class_name = self._get_singular_classname()
                if class_name is None:
                    raise TrestleError(
                        f'wrapper_alias not found for a collection type object: {elem.__class__.__name__}'
                    )
            wrapper_alias = str_utils.classname_to_alias(class_name, AliasMode.JSON)

        self._wrapper_alias: str = wrapper_alias

    def _get_singular_classname(self) -> str:
        """Get the inner class name for list or dict objects."""
        # this assumes all items in list and all values in dict are same type
        class_name = None
        root = getattr(self._elem, '__root__', None)
        if root is not None:
            type_str = root.__class__.__name__
            if type_str == 'list':
                class_name = self._elem.__root__[0].__class__.__name__
            elif type_str == 'dict':
                class_name = list(self._elem.__root__.values())[0].__class__.__name__
        return class_name

    def get(self) -> OscalBaseModel:
        """Return the model object."""
        return self._elem

    def _split_element_path(self, element_path: ElementPath):
        """Split the element path into root_model and remaing attr names."""
        path_parts = element_path.get()
        root_model = path_parts[0]
        path_parts = path_parts[1:]

        return root_model, path_parts

    def get_at(self,
               element_path: ElementPath = None,
               check_parent: bool = True) -> Union[OscalBaseModel, List[OscalBaseModel]]:
        """Get the element at the specified element path.

        it will return the sub-model object at the path. Sub-model object
        can be of type OscalBaseModel or List
        """
        if element_path is None:
            return self._elem

        # find the root-model and element path parts
        _, path_parts = self._split_element_path(element_path)

        # TODO validate that self._elem is of same type as root_model

        # initialize the starting element for search
        elm = self._elem
        if hasattr(elm, '__root__') and (isinstance(elm.__root__, dict) or isinstance(elm.__root__, list)):
            elm = elm.__root__

        # if parent exists and does not end with wildcard, use the parent as the starting element for search
        if check_parent and element_path.get_parent(
        ) is not None and element_path.get_parent().get_last() != ElementPath.WILDCARD:
            elm_at = self.get_at(element_path.get_parent())
            if elm_at is None:
                raise TrestleNotFoundError(f'Invalid parent path {element_path.get_parent()}')
            elm = elm_at

        # return the sub-element at the specified path
        for attr in path_parts:
            if elm is None:
                break

            # process for wildcard and array indexes

            if attr == ElementPath.WILDCARD:
                break
            elif attr.isnumeric():
                if isinstance(elm, list):
                    elm = elm[int(attr)]
                else:
                    # index to a non list type should return None
                    return None
            else:
                elm = elm.get_field_value_by_alias(attr)

        return elm

    def get_preceding_element(self, element_path: ElementPath) -> Optional[OscalBaseModel]:
        """Get the preceding element in the path."""
        preceding_path = element_path.get_preceding_path()
        preceding_elm: Optional[OscalBaseModel] = self.get_at(preceding_path)
        return preceding_elm

    def _get_sub_element_obj(self, sub_element):
        """Convert sub element into allowed model obj."""
        if not self.is_allowed_sub_element_type(sub_element):
            raise TrestleError(
                f'Sub element must be one of "{self.get_allowed_sub_element_types()}", found "{sub_element.__class__}"'
            )

        model_obj = sub_element
        if isinstance(sub_element, Element):
            model_obj = sub_element.get()

        return model_obj

    def set_at(self, element_path: ElementPath, sub_element: OscalBaseModel) -> 'Element':
        """Set a sub_element at the path in the current element.

        Sub element can be Element, OscalBaseModel, list or None type
        It returns the element itself so that chaining operation can be done such as
            `element.set_at(path, sub-element).get()`.
        """
        # convert the element_path to ElementPath if needed
        if isinstance(element_path, str):
            element_path = ElementPath(element_path)

        # convert sub-element to OscalBaseModel if needed
        model_obj = self._get_sub_element_obj(sub_element)

        # find the root-model and element path parts
        _, path_parts = self._split_element_path(element_path)

        # TODO validate that self._elem is of same type as root_model

        # If wildcard is present, check the input type and determine the preceding element
        if element_path.get_last() == ElementPath.WILDCARD:
            # validate the type is either list or OscalBaseModel
            if not isinstance(model_obj, list) and not isinstance(model_obj, OscalBaseModel):
                raise TrestleError(
                    f'The model object needs to be a List or OscalBaseModel for path with "{ElementPath.WILDCARD}"'
                )

            # since wildcard * is there, we need to go one level up for preceding element in the path
            preceding_elm = self.get_preceding_element(element_path.get_preceding_path())
        else:
            # get the preceding element in the path
            preceding_elm = self.get_preceding_element(element_path)

        if preceding_elm is None:
            raise TrestleError(f'Invalid sub element path {element_path} with no valid preceding element')

        # check if it can be a valid sub_element of the parent
        sub_element_name = element_path.get_element_name().replace('-', '_')
        if hasattr(preceding_elm, sub_element_name) is False:
            raise TrestleError(
                f'Element "{preceding_elm.__class__}" does not have the attribute "{sub_element_name}" '
                f'of type "{model_obj.__class__}"'
            )

        # set the sub-element
        try:
            setattr(preceding_elm, sub_element_name, model_obj)
        except ValidationError:
            sub_element_class = self.get_sub_element_class(preceding_elm, sub_element_name)
            raise TrestleError(
                f'Validation error: {sub_element_name} is expected to be "{sub_element_class}", '
                f'but found "{model_obj.__class__}"'
            )

        # returning self will allow to do 'chaining' of commands after set
        return self

    def to_yaml(self) -> str:
        """Convert into YAML string."""
        yaml = YAML(typ='safe')
        yaml.default_flow_style = False
        from io import StringIO
        string_stream = StringIO()
        yaml.dump(yaml.load(self.to_json(pretty=False)), string_stream)
        yaml_data = string_stream.getvalue()
        string_stream.close()

        return yaml_data

    def to_json(self, pretty: bool = True) -> str:
        """Convert into JSON string."""
        if self._wrapper_alias == self.IGNORE_WRAPPER_ALIAS:
            json_data = self._elem.oscal_serialize_json(pretty=pretty, wrapped=False)

        else:
            # Note before trying to edit this
            # This transient model allows self._elem not be an OscalBaseModel (e.g. a DICT or LIST)
            # typing need to be clarified.
            if isinstance(self._elem, OscalBaseModel):
                json_data = self._elem.oscal_serialize_json(pretty=pretty)
            else:
                dynamic_passer = {}
                dynamic_passer['TransientField'] = (self._elem.__class__, Field(self, alias=self._wrapper_alias))
                wrapper_model = create_model('TransientModel', __base__=OscalBaseModel, **dynamic_passer)
                wrapped_model = wrapper_model.construct(**{self._wrapper_alias: self._elem})
                json_data = wrapped_model.oscal_serialize_json(pretty=pretty, wrapped=False)
        return json_data

    @classmethod
    def get_sub_element_class(cls, parent_elm: OscalBaseModel, sub_element_name: str):
        """Get the class of the sub-element."""
        sub_element_class = parent_elm.__fields__[sub_element_name].outer_type_
        return sub_element_class

    @classmethod
    def get_allowed_sub_element_types(cls) -> List[str]:
        """Get the list of allowed sub element types."""
        return cls._allowed_sub_element_types

    @classmethod
    def is_allowed_sub_element_type(cls, elm) -> bool:
        """Check if is of allowed sub element type."""
        # FIXME: The following logic does not use the _allowed_sub_element_types being defined for the class
        if (isinstance(elm, Element) or isinstance(elm, OscalBaseModel) or isinstance(elm, list)
                or isinstance(elm, dict) or elm is None):
            return True

        return False

    def __str__(self) -> str:
        """Return string representation of element."""
        return type(self._elem).__name__

    def __eq__(self, other: object) -> bool:
        """Check that two elements are equal."""
        if not isinstance(other, Element):
            return False

        return self.get() == other.get()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle tasks base templating."""
import configparser
import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional

logger = logging.getLogger(__name__)


class TaskOutcome(Enum):
    """Enum describing possible task outcomes."""

    SUCCESS = 'success'
    FAILURE = 'failure'
    ROLLEDBACK = 'rolledback'
    SIM_SUCCESS = 'simulated-success'
    SIM_FAILURE = 'simulated-failure'
    NOT_IMPLEMENTED = 'not-implemented'


class TaskBase(ABC):
    """
    Abstract base class for tasks.

    Attributes:
        name: Name of the task.
    """

    name: str = 'base'

    def __init__(self, config_object: Optional[configparser.SectionProxy]) -> None:
        """Initialize task base and store config."""
        self._config = config_object

    @abstractmethod
    def print_info(self) -> None:
        """Print the help string."""

    @abstractmethod
    def execute(self) -> TaskOutcome:
        """Execute the task including potential rollback."""

    @abstractmethod
    def simulate(self) -> TaskOutcome:
        """Simulate the task and report task outcome."""


class PassFail(TaskBase):
    """
    Holding pattern template for a task which does nothing and always passes.

    Attributes:
        name: Name of the task.
    """

    name = 'pass-fail'

    def __init__(self, config_object: Optional[configparser.SectionProxy]) -> None:
        """
        Initialize trestle task pass-fail.

        Attributes:
            config_object: Config section associated with the task.
        """
        super().__init__(config_object)

    def print_info(self) -> None:
        """Print the help string."""
        logger.info(f'Help information for {self.name} task.')
        logger.info('This is a template task which reports pass fail depending on the specific configuration.')
        logger.info(
            'In this case if no config section is provided the task will fail. This a a task specific behavior.'
        )
        logger.info('Configuration flags sit under [task.pass-fail]')
        logger.info('with two boolean flags')
        logger.info('execute_status = True/False with a default pass')
        logger.info('simulate_status = True/False with a default fail')
        logger.info('Note that if the config file does not have the appropriate section this should fail.')
        logger.info('The princple goal is a simple development example.')

    def simulate(self) -> TaskOutcome:
        """Provide a simulated outcome."""
        if self._config:
            outcome = self._config.getboolean('simulate_status', fallback=True)
            if outcome:
                return TaskOutcome('simulated-success')
        return TaskOutcome('simulated-failure')

    def execute(self) -> TaskOutcome:
        """Provide a actual outcome."""
        if self._config:
            outcome = self._config.getboolean('execute_status', fallback=True)
            if outcome:
                return TaskOutcome('success')
        return TaskOutcome('failure')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle author folders subcommand."""
import pathlib
import shutil
import sys
from uuid import uuid4

from _pytest.monkeypatch import MonkeyPatch

import pytest

from tests import test_utils

import trestle.cli
from trestle.common import file_utils
from trestle.core.commands.author.consts import START_TEMPLATE_VERSION


@pytest.mark.parametrize(
    'command_string, return_code',
    [
        ('trestle author folders setup', 2), ('trestle author folders setup --task-name foobaa', 0),
        ('trestle author folders setup --task-name catalogs', 1)
    ]
)
def test_governed_folders_high(
    tmp_trestle_dir: pathlib.Path, command_string: str, return_code: int, monkeypatch: MonkeyPatch
) -> None:
    """Simple execution tests of trestle author folders."""
    monkeypatch.setattr(sys, 'argv', command_string.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
        # FIXME: Needs to be changed once implemented.
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == return_code


@pytest.mark.parametrize(
    'task_name, template_content, target_content, setup_code, template_code, validate_code, readme_validate',
    [
        (
            'test_task',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance'),
            0,
            0,
            0,
            False
        ),
        (
            'catalogs',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance'),
            1,
            1,
            1,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance_limits_of_changes'),
            0,
            0,
            0,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/bad_instance_renamed'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/bad_instance_mixed_name'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/bad_instance_missing'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/bad_instance_bad_content'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task_1',
            pathlib.Path('author/governed_folders/utf16test_good'),
            pathlib.Path('author/governed_folders/utf16test_bad'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task_2',
            pathlib.Path('author/governed_folders/utf16test_bad'),
            pathlib.Path('author/governed_folders/utf16test_good'),
            0,
            1,
            1,
            False
        ),
        (
            'another_test_task_3',
            pathlib.Path('author/governed_folders/template_with_readme'),
            pathlib.Path('author/governed_folders/good_instance_with_readme'),
            0,
            1,
            1,
            False
        ),
        (
            'another_test_task_4',
            pathlib.Path('author/governed_folders/template_with_readme'),
            pathlib.Path('author/governed_folders/good_instance_with_readme'),
            0,
            0,
            0,
            True
        ),
        (
            'another_test_task_5',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance_with_readme'),
            0,
            0,
            0,
            False
        ),
        # Note this will pass as templates are permissive to allow extra files
        # (e.g. it validates what people want to be validated and not supplementary assets)
        (
            'another_test_task_6',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance_with_readme'),
            0,
            0,
            0,
            True
        ),
        (
            'another_test_task_7',
            pathlib.Path('author/governed_folders/template_with_readme'),
            pathlib.Path('author/governed_folders/good_instance'),
            0,
            0,
            1,
            True
        ),
        (
            'another_test_task_8',
            pathlib.Path('author/governed_folders/template_with_readme'),
            pathlib.Path('author/governed_folders/good_instance'),
            0,
            1,
            1,
            False
        ),
        (
            'another_test_task_9',
            pathlib.Path('author/governed_folders/template_folder_with_drawio'),
            pathlib.Path('author/governed_folders/bad_folder_is_a_file'),
            0,
            0,
            0,  # this passes as random extra files are allowed there. This should be looked at.
            True
        ),
        (
            'another_test_task_10',
            pathlib.Path('author/governed_folders/bad_folder_is_a_file'),
            pathlib.Path('author/governed_folders/good_instance'),
            0,
            1,
            1,
            True
        ),
        (
            'another_test_task_11',
            pathlib.Path('author/governed_folders/template_folder_with_drawio'),
            pathlib.Path('author/governed_folders/template_folder_with_drawio'),
            0,
            0,
            0,
            True
        ),
        (
            'another_test_task_12',
            pathlib.Path('author/governed_folders/template_folder_with_drawio'),
            pathlib.Path('author/governed_folders/folder_with_bad_drawio'),
            0,
            0,
            1,
            True
        )
    ]
)
def test_e2e(
    task_name: str,
    template_content: pathlib.Path,
    target_content: pathlib.Path,
    setup_code: int,
    template_code: int,
    validate_code: int,
    readme_validate: bool,
    testdata_dir: pathlib.Path,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Run an E2E workflow with two test criteria for success."""
    # Note testdata_dir must be before tmp_trestle_dir in the argument order.
    readme_flag = '-rv' if readme_validate else ''
    command_string_setup = f'trestle author folders setup -tn {task_name} {readme_flag}'
    command_string_create_sample = f'trestle author folders create-sample -tn {task_name} {readme_flag}'
    command_string_validate_template = f'trestle author folders template-validate -tn {task_name} {readme_flag}'
    command_string_validate_content = f'trestle author folders validate -tn {task_name} --header-validate {readme_flag}'
    template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / task_name / START_TEMPLATE_VERSION
    test_content_loc = tmp_trestle_dir / task_name / f'{uuid4()}'
    if template_content == pathlib.Path('author/governed_folders/template_folder_with_drawio'):
        hidden_file = testdata_dir / pathlib.Path(
            'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
        )
        test_utils.make_file_hidden(hidden_file)
    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    rc = trestle.cli.Trestle().run()
    assert rc == setup_code
    if setup_code > 0:
        return
    shutil.rmtree(str(template_target_loc))
    template_loc = testdata_dir / template_content
    test_utils.copy_tree_or_file_with_hidden(template_loc, template_target_loc)

    monkeypatch.setattr(sys, 'argv', command_string_validate_template.split())
    rc = trestle.cli.Trestle().run()
    assert rc == template_code
    if template_code > 0:
        return

    # Create sample - should always work if we are here.
    monkeypatch.setattr(sys, 'argv', command_string_create_sample.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0

    target_source = testdata_dir / target_content
    test_utils.copy_tree_or_file_with_hidden(target_source, test_content_loc)

    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == validate_code


@pytest.mark.parametrize(
    'task_name, template_content, target_content, setup_code, template_code, validate_code, readme_validate',
    [
        (
            'test_task',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance'),
            0,
            0,
            0,
            False
        ),
        (
            'catalogs',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance'),
            1,
            1,
            1,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance_limits_of_changes'),
            0,
            0,
            0,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/bad_instance_renamed'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/bad_instance_mixed_name'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/bad_instance_missing'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task/with_a_sub_dir',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/bad_instance_bad_content'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task_1',
            pathlib.Path('author/governed_folders/utf16test_good'),
            pathlib.Path('author/governed_folders/utf16test_bad'),
            0,
            0,
            1,
            False
        ),
        (
            'another_test_task_2',
            pathlib.Path('author/governed_folders/utf16test_bad'),
            pathlib.Path('author/governed_folders/utf16test_good'),
            0,
            1,
            1,
            False
        ),
        (
            'another_test_task_3',
            pathlib.Path('author/governed_folders/template_with_readme'),
            pathlib.Path('author/governed_folders/good_instance_with_readme'),
            0,
            1,
            1,
            False
        ),
        (
            'another_test_task_4',
            pathlib.Path('author/governed_folders/template_with_readme'),
            pathlib.Path('author/governed_folders/good_instance_with_readme'),
            0,
            0,
            0,
            True
        ),
        (
            'another_test_task_5',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance_with_readme'),
            0,
            0,
            0,
            False
        ),
        # Note this will pass as templates are permissive to allow extra files
        # (e.g. it validates what people want to be validated and not supplementary assets)
        (
            'another_test_task_6',
            pathlib.Path('author/governed_folders/template_folder'),
            pathlib.Path('author/governed_folders/good_instance_with_readme'),
            0,
            0,
            0,
            True
        ),
        (
            'another_test_task_7',
            pathlib.Path('author/governed_folders/template_with_readme'),
            pathlib.Path('author/governed_folders/good_instance'),
            0,
            0,
            1,
            True
        ),
        (
            'another_test_task_8',
            pathlib.Path('author/governed_folders/template_with_readme'),
            pathlib.Path('author/governed_folders/good_instance'),
            0,
            1,
            1,
            False
        ),
        (
            'another_test_task_9',
            pathlib.Path('author/governed_folders/template_folder_with_drawio'),
            pathlib.Path('author/governed_folders/bad_folder_is_a_file'),
            0,
            0,
            0,  # this passes as random extra files are allowed there. This should be looked at.
            True
        ),
        (
            'another_test_task_10',
            pathlib.Path('author/governed_folders/bad_folder_is_a_file'),
            pathlib.Path('author/governed_folders/good_instance'),
            0,
            1,
            1,
            True
        ),
        (
            'another_test_task_11',
            pathlib.Path('author/governed_folders/template_folder_with_drawio'),
            pathlib.Path('author/governed_folders/template_folder_with_drawio'),
            0,
            0,
            0,
            True
        ),
        (
            'another_test_task_12',
            pathlib.Path('author/governed_folders/template_folder_with_drawio'),
            pathlib.Path('author/governed_folders/folder_with_bad_drawio'),
            0,
            0,
            1,
            True
        )
    ]
)
def test_e2e_backward_compatibility(
    task_name: str,
    template_content: pathlib.Path,
    target_content: pathlib.Path,
    setup_code: int,
    template_code: int,
    validate_code: int,
    readme_validate: bool,
    testdata_dir: pathlib.Path,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Run same E2E workflow but pretend that there are existing templates."""
    # Note testdata_dir must be before tmp_trestle_dir in the argument order.
    readme_flag = '-rv' if readme_validate else ''
    command_string_setup = f'trestle author folders setup -tn {task_name} {readme_flag}'
    command_string_create_sample = f'trestle author folders create-sample -tn {task_name} {readme_flag}'
    command_string_validate_template = f'trestle author folders template-validate -tn {task_name} {readme_flag}'
    command_string_validate_content = f'trestle author folders validate -tn {task_name} --header-validate {readme_flag}'
    template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / task_name / START_TEMPLATE_VERSION
    old_template_path = tmp_trestle_dir / '.trestle' / 'author' / task_name
    test_content_loc = tmp_trestle_dir / task_name / f'{uuid4()}'

    if template_content == pathlib.Path('author/governed_folders/template_folder_with_drawio'):
        hidden_file = testdata_dir / pathlib.Path(
            'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
        )
        test_utils.make_file_hidden(hidden_file)

    # Add old templates
    template_loc = testdata_dir / template_content
    test_utils.copy_tree_or_file_with_hidden(template_loc, old_template_path)

    assert old_template_path.exists()
    assert not template_target_loc.exists()

    # Setup, make sure filesystem is updated.
    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    rc = trestle.cli.Trestle().run()
    assert rc == setup_code
    if setup_code > 0:
        return

    all_files_wo_version = list(
        filter(lambda p: p.is_file() and not file_utils.is_hidden(p), (old_template_path.iterdir()))
    )
    all_files_w_version = list(
        filter(lambda p: p.is_file() and not file_utils.is_hidden(p), (template_target_loc.iterdir()))
    )

    if template_loc.is_dir():
        all_old_files = list(filter(lambda p: p.is_file() and not file_utils.is_hidden(p), (template_loc.iterdir())))
        all_old_files = [el.parts[-1] for el in all_old_files]
    else:
        all_old_files = [template_loc.parts[-1]]

    assert len(all_files_wo_version) == 0
    assert template_target_loc.exists()
    assert all(el in [el.parts[-1] for el in all_files_w_version] for el in all_old_files)

    shutil.rmtree(str(template_target_loc))
    template_loc = testdata_dir / template_content
    test_utils.copy_tree_or_file_with_hidden(template_loc, template_target_loc)

    monkeypatch.setattr(sys, 'argv', command_string_validate_template.split())
    rc = trestle.cli.Trestle().run()
    assert rc == template_code
    if template_code > 0:
        return

    # Create sample - should always work if we are here.
    monkeypatch.setattr(sys, 'argv', command_string_create_sample.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0

    target_source = testdata_dir / target_content
    if target_source.is_dir():
        shutil.copytree(str(target_source), str(test_content_loc))
    else:
        shutil.copy2(str(target_source), str(test_content_loc))

    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == validate_code


def test_ignore_flag(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that ignored files are not validated. Validation will fail if attempted."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/governed_folders/template_folder_with_drawio'
    test_instances_folder = testdata_dir / 'author/governed_folders/ignored_files'
    task_instance_folder = tmp_trestle_dir / 'test_task/folder_1'

    hidden_file = testdata_dir / pathlib.Path(
        'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
    )
    test_utils.make_file_hidden(hidden_file)

    test_utils.copy_tree_or_file_with_hidden(test_template_folder, task_template_folder)

    # copy all files
    shutil.copytree(test_instances_folder, task_instance_folder)

    command_string_validate_content = 'trestle author folders validate -tn test_task -ig ^_.*'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0


def test_fail_when_no_task(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that validation fails when task does not exists."""
    command_validate = 'trestle author folders validate -tn test_task'
    monkeypatch.setattr(sys, 'argv', command_validate.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 1


def test_passes_when_extra_md_file(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test extra file in the correct folder."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/governed_folders/template_folder'
    test_instances_folder = testdata_dir / 'author/governed_folders/good_instance_with_readme'
    task_instance_folder = tmp_trestle_dir / 'test_task/folder_1'

    hidden_file = testdata_dir / pathlib.Path(
        'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
    )
    test_utils.make_file_hidden(hidden_file)

    test_utils.copy_tree_or_file_with_hidden(test_template_folder, task_template_folder)

    shutil.copytree(test_instances_folder, task_instance_folder)

    # Add that extra file
    (task_instance_folder / 'extra_document.md').touch()

    command_string_validate_content = 'trestle author folders validate -tn test_task -hv'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0

    command_string_validate_content = 'trestle author folders validate -tn test_task'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0


def test_instance_no_header(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test behaviour when validating instance with no header."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/governed_folders/template_folder'
    test_instances_folder = testdata_dir / 'author/governed_folders/bad_instance_no_header'
    task_instance_folder = tmp_trestle_dir / 'test_task/folder_1'

    hidden_file = testdata_dir / pathlib.Path(
        'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
    )
    test_utils.make_file_hidden(hidden_file)

    test_utils.copy_tree_or_file_with_hidden(test_template_folder, task_template_folder)

    shutil.copytree(test_instances_folder, task_instance_folder)

    command_string_validate_content = 'trestle author folders validate -tn test_task -hv'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 1

    command_string_validate_content = 'trestle author folders validate -tn test_task'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0

    command_string_validate_content = 'trestle author folders validate -tn test_task -tv 0.0.1'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0

    command_string_validate_content = 'trestle author folders validate -tn test_task -tv 0.0.1 -hv'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 1


def test_drawio_versioning_validation(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test behaviour when validating drawio instance."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/governed_folders/template_folder_with_version'
    test_instances_folder = testdata_dir / 'author/governed_folders/good_instance_with_version'
    task_instance_folder = tmp_trestle_dir / 'test_task/folder_1'

    hidden_file = testdata_dir / pathlib.Path(
        'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
    )
    test_utils.make_file_hidden(hidden_file)

    test_utils.copy_tree_or_file_with_hidden(test_template_folder, task_template_folder)

    shutil.copytree(test_instances_folder, task_instance_folder)

    command_string_validate_content = 'trestle author folders validate -tn test_task -hv'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0

    command_string_validate_content = 'trestle author folders validate -tn test_task -hv -tv 0.0.1'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0


def test_validate_template_with_type_field(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test behaviour when validating an instance with x-trestle-template-type field."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/governed_folders/template_folder_with_template_type'
    test_instances_folder = testdata_dir / 'author/governed_folders/good_instance_with_template_type'
    task_instance_folder = tmp_trestle_dir / 'test_task/folder_1'

    hidden_file = testdata_dir / pathlib.Path(
        'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
    )
    test_utils.make_file_hidden(hidden_file)

    test_utils.copy_tree_or_file_with_hidden(test_template_folder, task_template_folder)

    shutil.copytree(test_instances_folder, task_instance_folder)
    # test validate short
    command_string_validate_content = 'trestle author folders validate -tn test_task -hv -vtt'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0

    # test validate long
    command_string_validate_content = 'trestle author folders validate -tn test_task -hv --validate-template-type'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0


def test_validate_template_with_type_field_unhappy(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test hunhappy behaviour when validating an instance with x-trestle-template-type field."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/governed_folders/template_folder_with_template_type'
    test_instances_folder = testdata_dir / 'author/governed_folders/good_instance_without_template_type'
    task_instance_folder = tmp_trestle_dir / 'test_task/folder_1'

    hidden_file = testdata_dir / pathlib.Path(
        'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
    )
    test_utils.make_file_hidden(hidden_file)

    test_utils.copy_tree_or_file_with_hidden(test_template_folder, task_template_folder)

    shutil.copytree(test_instances_folder, task_instance_folder)
    # test validate short
    command_string_validate_content = 'trestle author folders validate -tn test_task -hv -vtt'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 1

    # test validate long
    command_string_validate_content = 'trestle author folders validate -tn test_task -hv --validate-template-type'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 1


def test_heading_levels_hierarchy(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test behaviour when validating drawio instance."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    test_template_folder = testdata_dir / 'author/governed_folders/template_folder_headling_levels'
    test_instances_folder = testdata_dir / 'author/governed_folders/instance_with_diff_heading_levels'
    task_instance_folder = tmp_trestle_dir / 'test_task/folder_1'

    hidden_file = testdata_dir / pathlib.Path(
        'author/governed_folders/template_folder_with_drawio/.hidden_does_not_affect'
    )
    test_utils.make_file_hidden(hidden_file)

    test_utils.copy_tree_or_file_with_hidden(test_template_folder, task_template_folder)

    shutil.copytree(test_instances_folder, task_instance_folder)

    command_string_validate_content = 'trestle author folders validate -tn test_task -hv'
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Commands."""
import argparse
import logging
import operator
import pathlib
import re
import uuid
from typing import Any, Dict, Optional

from jinja2 import ChoiceLoader, DictLoader, Environment, FileSystemLoader, Template

from ruamel.yaml import YAML

from trestle.common import const, log
from trestle.common.err import TrestleIncorrectArgsError, handle_generic_command_exception
from trestle.common.load_validate import load_validate_model_name
from trestle.common.model_utils import ModelUtils
from trestle.core.catalog.catalog_interface import CatalogInterface
from trestle.core.commands.command_docs import CommandPlusDocs
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.control_interface import ControlInterface, ParameterRep
from trestle.core.docs_control_writer import DocsControlWriter
from trestle.core.jinja import MDCleanInclude, MDDatestamp, MDSectionInclude
from trestle.core.profile_resolver import ProfileResolver
from trestle.core.ssp_io import SSPMarkdownWriter
from trestle.oscal.profile import Profile
from trestle.oscal.ssp import SystemSecurityPlan

logger = logging.getLogger(__name__)


class JinjaCmd(CommandPlusDocs):
    """Transform an input template to an output document using jinja templating."""

    max_recursion_depth = 2

    name = 'jinja'

    def _init_arguments(self) -> None:
        self.add_argument('-i', '--input', help='Input jinja template, relative to trestle root', required=True)
        self.add_argument('-o', '--output', help='Output template, relative to trestle root.', required=True)
        self.add_argument(
            '-lut',
            '--look-up-table',
            help='Key-value pair table, stored as yaml, to be passed to jinja as variables',
            required=False
        )
        self.add_argument(
            '-elp',
            '--external-lut-prefix',
            help='Prefix paths for LUT, to maintain compatibility with other templating systems',
            required=False
        )
        self.add_argument(
            '-nc',
            '--number-captions',
            help='Add incremental numbering to table and image captions, in the form Table n - ... and Figure n - ...',
            action='store_true'
        )
        self.add_argument(
            '-bf',
            '--bracket-format',
            help='With -sv, allows brackets around value, e.g. [.] or ((.)), with the dot representing the value.',
            required=False
        )
        self.add_argument(
            '-vap',
            '--value-assigned-prefix',
            help='Places a prefix in front of the parameter string if a value has been assigned.',
            required=False,
            type=str,
            default=''
        )
        self.add_argument(
            '-vnap',
            '--value-not-assigned-prefix',
            help='Places a prefix in front of the parameter string if a value has *not* been assigned.',
            required=False,
            type=str,
            default=''
        )
        self.add_argument(
            '-ssp', '--system-security-plan', help='An optional SSP to be passed', default=None, required=False
        )
        self.add_argument('-p', '--profile', help='An optional profile to be passed', default=None, required=False)
        self.add_argument(
            '-dp',
            '--docs-profile',
            help='Output profile controls to separate markdown files',
            action='store_true',
            required=False
        )

    def _run(self, args: argparse.Namespace) -> None:
        try:
            log.set_log_level_from_args(args)
            logger.debug(f'Starting {self.name} command')
            input_path = pathlib.Path(args.input)
            output_path = pathlib.Path(args.output)

            logger.debug(f'input_path: {input_path}')
            logger.debug(f'output_path: {output_path}')
            logger.debug(f'system_security_plan path: {args.system_security_plan}')
            logger.debug(f'profile: {args.profile}')
            logger.debug(f'docs_profile: {args.docs_profile}')
            logger.debug(f'lookup_table: {args.look_up_table}')

            if args.system_security_plan and args.docs_profile:
                raise TrestleIncorrectArgsError('Output to multiple files is possible with profile only.')

            if args.docs_profile and not args.profile:
                raise TrestleIncorrectArgsError('Profile must be provided to output to multiple files.')

            lut = {}
            if args.look_up_table:
                lut_table = pathlib.Path(args.look_up_table)
                lookup_table_path = pathlib.Path.cwd() / lut_table
                lut = JinjaCmd.load_LUT(lookup_table_path, args.external_lut_prefix)

            if args.profile and args.docs_profile:
                return JinjaCmd.jinja_multiple_md(
                    pathlib.Path(args.trestle_root),
                    input_path,
                    output_path,
                    args.profile,
                    lut,
                    parameters_formatting=args.bracket_format,
                    value_assigned_prefix=args.value_assigned_prefix,
                    value_not_assigned_prefix=args.value_not_assigned_prefix
                )

            return JinjaCmd.jinja_ify(
                pathlib.Path(args.trestle_root),
                input_path,
                output_path,
                args.system_security_plan,
                args.profile,
                lut,
                number_captions=args.number_captions,
                parameters_formatting=args.bracket_format,
                value_assigned_prefix=args.value_assigned_prefix,
                value_not_assigned_prefix=args.value_not_assigned_prefix
            )

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(  # type: ignore
                e, logger, 'Error while generating markdown via Jinja template'
            )

    @staticmethod
    def load_LUT(path: pathlib.Path, prefix: Optional[str]) -> Dict[str, Any]:  # noqa: N802
        """Load a Yaml lookup table from file."""
        yaml = YAML()
        lut = yaml.load(path.open('r', encoding=const.FILE_ENCODING))
        if prefix:
            prefixes = prefix.split('.')
            while prefixes:
                old_lut = lut
                lut[prefixes.pop(-1)] = old_lut

        return lut

    @staticmethod
    def jinja_ify(
        trestle_root: pathlib.Path,
        r_input_file: pathlib.Path,
        r_output_file: pathlib.Path,
        ssp: Optional[str],
        profile: Optional[str],
        lut: Dict[str, Any],
        number_captions: Optional[bool] = False,
        parameters_formatting: Optional[str] = None,
        value_assigned_prefix: Optional[str] = None,
        value_not_assigned_prefix: Optional[str] = None
    ) -> int:
        """Run jinja over an input file with additional booleans."""
        template_folder = pathlib.Path.cwd()
        jinja_env = Environment(
            loader=FileSystemLoader(template_folder),
            extensions=[MDSectionInclude, MDCleanInclude, MDDatestamp],
            trim_blocks=True,
            autoescape=True
        )
        template = jinja_env.get_template(str(r_input_file))
        # create boolean dict
        if operator.xor(bool(ssp), bool(profile)):
            raise TrestleIncorrectArgsError('Both SSP and profile should be provided or not at all')

        if ssp:
            # name lookup
            ssp_data, _ = load_validate_model_name(trestle_root, ssp, SystemSecurityPlan)
            lut['ssp'] = ssp_data
            profile_path = ModelUtils.get_model_path_for_name_and_class(trestle_root, profile, Profile)
            profile_resolver = ProfileResolver()
            resolved_catalog = profile_resolver.get_resolved_profile_catalog(
                trestle_root,
                profile_path,
                False,
                False,
                parameters_formatting,
                ParameterRep.ASSIGNMENT_FORM,
                False,
                value_assigned_prefix,
                value_not_assigned_prefix
            )

            ssp_writer = SSPMarkdownWriter(trestle_root)
            ssp_writer.set_ssp(ssp_data)
            ssp_writer.set_catalog(resolved_catalog)
            lut['catalog'] = resolved_catalog
            lut['catalog_interface'] = CatalogInterface(resolved_catalog)
            lut['control_interface'] = ControlInterface()
            lut['control_writer'] = DocsControlWriter()
            lut['ssp_md_writer'] = ssp_writer

        output = JinjaCmd.render_template(template, lut, template_folder)

        output_file = trestle_root / r_output_file
        if number_captions:
            output_file.open('w', encoding=const.FILE_ENCODING).write(_number_captions(output))
        else:
            output_file.open('w', encoding=const.FILE_ENCODING).write(output)

        return CmdReturnCodes.SUCCESS.value

    @staticmethod
    def jinja_multiple_md(
        trestle_root: pathlib.Path,
        r_input_file: pathlib.Path,
        r_output_file: pathlib.Path,
        profile_name: Optional[str],
        lut: Dict[str, Any],
        parameters_formatting: Optional[str] = None,
        value_assigned_prefix: Optional[str] = None,
        value_not_assigned_prefix: Optional[str] = None
    ) -> int:
        """Output profile as multiple markdown files using Jinja."""
        template_folder = pathlib.Path.cwd()

        # Output to multiple markdown files
        profile, profile_path = ModelUtils.load_model_for_class(trestle_root, profile_name, Profile)
        profile_resolver = ProfileResolver()
        resolved_catalog = profile_resolver.get_resolved_profile_catalog(
            trestle_root,
            profile_path,
            False,
            False,
            parameters_formatting,
            ParameterRep.ASSIGNMENT_FORM,
            False,
            value_assigned_prefix,
            value_not_assigned_prefix
        )
        catalog_interface = CatalogInterface(resolved_catalog)

        # Generate a single markdown page for each control per each group
        for group in catalog_interface.get_all_groups_from_catalog():
            for control in catalog_interface.get_sorted_controls_in_group(group.id):
                _, group_title, _ = catalog_interface.get_group_info_by_control(control.id)
                group_dir = r_output_file
                control_path = catalog_interface.get_control_path(control.id)
                for sub_dir in control_path:
                    group_dir = group_dir / sub_dir
                    if not group_dir.exists():
                        group_dir.mkdir(parents=True, exist_ok=True)

                control_writer = DocsControlWriter()

                jinja_env = Environment(
                    loader=FileSystemLoader(template_folder),
                    extensions=[MDSectionInclude, MDCleanInclude, MDDatestamp],
                    trim_blocks=True,
                    autoescape=True
                )
                template = jinja_env.get_template(str(r_input_file))
                lut['catalog_interface'] = catalog_interface
                lut['control_interface'] = ControlInterface()
                lut['control_writer'] = control_writer
                lut['control'] = control
                lut['profile'] = profile
                lut['group_title'] = group_title
                output = JinjaCmd.render_template(template, lut, template_folder)

                output_file = trestle_root / group_dir / pathlib.Path(control.id + const.MARKDOWN_FILE_EXT)
                output_file.open('w', encoding=const.FILE_ENCODING).write(output)

        return CmdReturnCodes.SUCCESS.value

    @staticmethod
    def render_template(template: Template, lut: Dict[str, Any], template_folder: pathlib.Path) -> str:
        """Render template."""
        new_output = template.render(**lut)
        output = ''
        # This recursion allows nesting within expressions (e.g. an expression can contain jinja templates).
        error_countdown = JinjaCmd.max_recursion_depth
        while new_output != output and error_countdown > 0:
            error_countdown = error_countdown - 1
            output = new_output
            random_name = uuid.uuid4()  # Should be random and not used.
            dict_loader = DictLoader({str(random_name): new_output})
            jinja_env = Environment(
                loader=ChoiceLoader([dict_loader, FileSystemLoader(template_folder)]),
                extensions=[MDCleanInclude, MDSectionInclude, MDDatestamp],
                autoescape=True,
                trim_blocks=True
            )
            template = jinja_env.get_template(str(random_name))
            new_output = template.render(**lut)

        return output


def _number_captions(md_body: str) -> str:
    """Incrementally number tables and image captions."""
    images = {}
    tables = {}
    output = md_body.splitlines()

    for index, line in enumerate(output):
        if re.match(r'!\[.+\]\(.+\)', line):
            images[index] = line
        if output[index].lower().startswith('table: '):
            tables[index] = line

    for index, row in enumerate(tables):
        output[row] = f'Table: Table {index + 1} - {tables[row].split(": ")[1]}'

    for index, row in enumerate(images):
        output[row] = images[row].replace('![', f'![Figure {index + 1} - ')

    return '\n'.join(output)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Validate command tests."""
import argparse
import configparser
import pathlib
import sys

from _pytest.monkeypatch import MonkeyPatch

import pytest

import trestle.cli
import trestle.common.const as const
import trestle.common.err as err
import trestle.core.commands.task as taskcmd
from trestle.tasks.base_task import PassFail


def test_get_list_cli(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Simple test which does E2E tests."""
    command = 'trestle task -l'

    monkeypatch.setattr(sys, 'argv', command.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == 0


def test_arguments(tmp_trestle_dir: pathlib.Path) -> None:
    """Test that the argument loading behave correctly."""
    # should pass and print a list
    # also should work but same as above
    # NOTE: When passing arguments by hand like this - they are defined if missing and passed as None
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, trestname='task', list=True, verbose=1, task=None, config=None, info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc == 0

    # Conflicting arguments: info an dlist
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, name='task', list=True, verbose=1, task='stuff', config=None, info=True
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc > 0

    # should not work - missing required arguments
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name='task',
        config='config_file.json',
        verbose=1,
        list=False,
        task=None,
        info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc > 0
    # Failure due to task not in task pool.
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name='task',
        list=False,
        verbose=1,
        task='this-is-not-a-task',
        config=False,
        info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc > 0


def test_load_default_config(tmp_trestle_dir: pathlib.Path) -> None:
    """Overwrite the default config file to ensure success. Note this is coupled to PassFail class."""
    # Load and overwrite config file
    section_name = 'task.pass-fail'
    config_object = configparser.ConfigParser()
    config_path = pathlib.Path(const.TRESTLE_CONFIG_DIR) / const.TRESTLE_CONFIG_FILE
    config_object.read(config_path)
    # add section
    config_object.add_section(section_name)
    config_object[section_name]['execute_status'] = 'True'
    config_object[section_name]['simulate_status'] = 'True'
    config_object.write(config_path.open('w', encoding=const.FILE_ENCODING))
    # Now good.

    # should print info.
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, name='task', list=False, verbose=1, task='pass-fail', config=None, info=True
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc == 0
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, name='task', list=False, verbose=1, task='pass-fail', config=None, info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc == 0
    #


def test_load_custom_config(tmp_trestle_dir: pathlib.Path) -> None:
    """Test functionality."""
    section_name = 'task.pass-fail'
    config_object = configparser.ConfigParser()
    config_path = pathlib.Path('config_file.json')
    config_object.read(config_path)
    # add section
    config_object.add_section(section_name)
    config_object[section_name]['execute_status'] = 'True'
    config_object[section_name]['simulate_status'] = 'True'
    config_object.write(config_path.open('w', encoding=const.FILE_ENCODING))

    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name='task',
        list=False,
        verbose=1,
        task='pass-fail',
        config='config_file.json',
        info=True
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc == 0

    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name='task',
        list=False,
        verbose=1,
        task='pass-fail',
        config='config_file.json',
        info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc == 0


def test_task_cmd_not_trestle_dir(tmp_empty_cwd: pathlib.Path) -> None:
    """Ensure that failure occurs when tasks is executed outside of a trestle dirctory."""
    section_name = 'task.pass-fail'
    config_object = configparser.ConfigParser()
    config_path = pathlib.Path('config_file.json')
    config_object.read(config_path)
    # add section
    config_object.add_section(section_name)
    config_object[section_name]['execute_status'] = 'True'
    config_object[section_name]['simulate_status'] = 'True'
    config_object.write(config_path.open('w', encoding=const.FILE_ENCODING))

    args = argparse.Namespace(
        trestle_root=tmp_empty_cwd,
        name='task',
        list=False,
        verbose=1,
        task='pass-fail',
        config='config_file.json',
        info=True
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc > 0


def test_task_cmd_incorrect_config_path_when_passed(tmp_trestle_dir: pathlib.Path) -> None:
    """Test failure occurs cleanly when a passed config file is not a config file."""
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir,
        name='task',
        list=False,
        verbose=1,
        task='pass-fail',
        config='made_up.json',
        info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc > 0


def test_pass_fail_respect_config(tmp_trestle_dir: pathlib.Path) -> None:
    """This tests whether pass-fail configuration is being successfully respected."""
    # Load and overwrite config file
    section_name = 'task.pass-fail'
    config_object = configparser.ConfigParser()
    config_path = pathlib.Path(const.TRESTLE_CONFIG_DIR) / const.TRESTLE_CONFIG_FILE
    config_object.read(config_path)
    # add section
    config_object.add_section(section_name)
    config_object[section_name]['execute_status'] = 'False'
    config_object[section_name]['simulate_status'] = 'True'
    config_object.write(config_path.open('w', encoding=const.FILE_ENCODING))
    # Now good.

    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, name='task', list=False, verbose=1, task='pass-fail', config=None, info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc > 0

    config_object['task.pass-fail']['execute_status'] = 'True'
    config_object['task.pass-fail']['simulate_status'] = 'False'
    config_object.write(config_path.open('w', encoding=const.FILE_ENCODING))
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, name='task', list=False, verbose=1, task='pass-fail', config=None, info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc > 0


def test_trestle_no_config(tmp_trestle_dir: pathlib.Path) -> None:
    """Warn but in this case fail when no config is provided."""
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, name='task', list=False, verbose=1, task='pass-fail', config=None, info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc == 1


def test_trestle_fail_on_task_exception(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Force an exception in a running trestle task and ensure it is captured."""

    def simulate_mock(*args, **kwargs):
        raise err.TrestleError('simulate_fail')

    # Load and overwrite config file
    section_name = 'task.pass-fail'
    config_object = configparser.ConfigParser()
    config_path = pathlib.Path(const.TRESTLE_CONFIG_DIR) / const.TRESTLE_CONFIG_FILE
    config_object.read(config_path)
    # add section
    config_object.add_section(section_name)
    config_object[section_name]['execute_status'] = 'True'
    config_object[section_name]['simulate_status'] = 'True'
    config_object.write(config_path.open('w', encoding=const.FILE_ENCODING))
    # Now good.
    monkeypatch.setattr(PassFail, 'execute', simulate_mock)
    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, name='task', list=False, verbose=1, task='pass-fail', config=None, info=False
    )
    rc = taskcmd.TaskCmd()._run(args)
    assert rc > 0



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provide a context for control operations."""
from __future__ import annotations

import copy
import pathlib
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

from trestle.common.list_utils import as_dict
from trestle.core.control_interface import CompDict
from trestle.oscal import component as comp
from trestle.oscal import profile as prof


class ContextPurpose(Enum):
    """Specify the modality of the control markdown."""

    CATALOG = 0
    COMPONENT = 1
    PROFILE = 2
    SSP = 3


@dataclass
class ControlContext:
    """Class encapsulating control markdown usage."""

    purpose: ContextPurpose
    to_markdown: bool
    trestle_root: pathlib.Path
    md_root: pathlib.Path
    prompt_responses: bool
    overwrite_header_values: bool
    set_parameters_flag: bool
    cli_yaml_header: Optional[Dict[Any, Any]] = None
    sections_dict: Optional[Dict[str, str]] = None
    profile: Optional[prof.Profile] = None
    required_sections: Optional[List[str]] = None
    allowed_sections: Optional[List[str]] = None
    comp_def: Optional[comp.ComponentDefinition] = None
    comp_name: Optional[str] = None
    component: Optional[comp.DefinedComponent] = None
    comp_def_name_list: Optional[List[str]] = None
    inherited_props: Optional[Dict[str, Any]] = None
    rules_dict: Optional[Dict[str, Dict[str, Any]]] = None
    rules_params_dict: Optional[Dict[str, Dict[str, Any]]] = None
    control_implementation: Optional[comp.ControlImplementation] = None
    uri_name_map: Optional[Dict[str, str]] = None
    comp_dict: Optional[CompDict] = None
    merged_header: Optional[Dict[str, Any]] = None

    @classmethod
    def generate(
        cls,
        purpose: ContextPurpose,
        to_markdown: bool,
        trestle_root: pathlib.Path,
        md_root: pathlib.Path,
        prompt_responses: bool = False,
        overwrite_header_values: bool = False,
        set_parameters_flag: bool = False,
        cli_yaml_header: Optional[Dict[Any, Any]] = None,
        sections_dict: Optional[Dict[str, str]] = None,
        profile: Optional[prof.Profile] = None,
        required_sections: Optional[List[str]] = None,
        allowed_sections: Optional[List[str]] = None,
        comp_def: Optional[comp.ComponentDefinition] = None,
        comp_name: Optional[str] = None,
        component: Optional[comp.DefinedComponent] = None,
        comp_def_name_list: Optional[List[str]] = None,
        inherited_props: Optional[Dict[str, Any]] = None,
        rules_dict: Optional[Dict[str, Dict[str, Any]]] = None,
        rules_params_dict: Optional[Dict[str, Dict[str, Any]]] = None,
        control_implementation: Optional[comp.ControlImplementation] = None,
        uri_name_map: Optional[Dict[str, str]] = None,
        comp_dict: Optional[CompDict] = None,
        merged_header: Optional[Dict[str, Any]] = None
    ) -> ControlContext:
        """Generate control context of the needed type."""
        context = cls(
            purpose,
            to_markdown,
            trestle_root,
            md_root,
            prompt_responses,
            overwrite_header_values,
            set_parameters_flag,
            cli_yaml_header=cli_yaml_header,
            sections_dict=sections_dict,
            profile=profile,
            required_sections=required_sections,
            allowed_sections=allowed_sections,
            comp_def=comp_def,
            comp_name=comp_name,
            component=component,
            comp_def_name_list=comp_def_name_list,
            inherited_props=inherited_props,
            rules_dict=rules_dict,
            rules_params_dict=rules_params_dict,
            control_implementation=control_implementation,
            uri_name_map=uri_name_map,
            comp_dict=comp_dict
        )
        context.cli_yaml_header = as_dict(cli_yaml_header)
        context.sections_dict = as_dict(sections_dict)
        context.merged_header = as_dict(merged_header)
        # catalog generate always sets params
        if to_markdown:
            context.set_parameters = True
        return context

    @classmethod
    def clone(cls, context: ControlContext) -> ControlContext:
        """Create a deep clone of the context without duplicating large objects."""
        new_context = cls(
            context.purpose,
            context.to_markdown,
            context.trestle_root,
            context.md_root,
            context.prompt_responses,
            context.overwrite_header_values,
            context.set_parameters_flag,
            cli_yaml_header=copy.deepcopy(context.cli_yaml_header),
            sections_dict=copy.deepcopy(context.sections_dict),
            profile=context.profile,
            required_sections=context.required_sections,
            allowed_sections=context.allowed_sections,
            comp_def=context.comp_def,
            comp_name=context.comp_name,
            component=context.component,
            comp_def_name_list=context.comp_def_name_list,
            inherited_props=copy.deepcopy(context.inherited_props),
            rules_dict=copy.deepcopy(context.rules_dict),
            rules_params_dict=copy.deepcopy(context.rules_params_dict),
            control_implementation=copy.deepcopy(context.control_implementation),
            uri_name_map=context.uri_name_map,
            comp_dict=copy.deepcopy(context.comp_dict),
            merged_header=copy.deepcopy(context.merged_header)
        )
        return new_context
